/*

	~~~~~~~~~~~
	~ gui.cpp ~
	~~~~~~~~~~~

Funktionen fr das GUI. Das GUI ist _kein_ Modul, sondern ein unabhngiger
Programmteil, der alle Module in einem groen Ganzen verwendet. Alle
Zeichenoperationen wurden mit FLTK und/oder OpenGL durchgefhrt.

Das GUI wurde erfolgreich kompiliert und getestet unter:
- Gentoo Linux
- FreeBSD

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/

#include <FL/Fl.h>
#include <FL/gl.h>
#include <FL/fl_ask.h>
#include <FL/Fl_Text_Display.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <sys/socket.h>

#include <vector>

#include "gui.h"
#include "protocol.h"
#include "sim_language.h"

using namespace std;

/// Konstruktor fr das Hauptfenster des GUI. Hier werden alle Steuerelemente
/// initialisiert sowie deren Callback-Funktionen gesetzt. Zudem fragt das GUI
/// beim Module Master an, ob alle bentigten Module verfgbar sind.
///
/// @param width Breite des Fensters
/// @param height Hhe des Fenster
/// @param title Titel des Fensters 
GUI_Window::GUI_Window( int width, int height, char* title ) : Fl_Window( width, height, title )
{
	// OpenGL initialisieren. Wird fr GUI_GL_Window bentigt. Hier lassen
	// sich fr schwachbrstigere Plattformen eventuell Tweaks anbringen.

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	// Steuerelemente zeichnen
	
	gl_win = new GUI_GL_Window( 10, 35, 320, 320, NULL );
	text_in = new Fl_Text_Editor( 10, 365, 125, 270 );
	text_out = new Fl_Text_Display( 145, 365, 520, 270 );

	text_in_buffer = new Fl_Text_Buffer;
	text_in->buffer( text_in_buffer );

	text_out_buffer = new Fl_Text_Buffer;
	text_out->buffer( text_out_buffer );
	
	text_in->textsize( 11 );
	text_out->textsize( 11 );
	
	text_in_buffer->text( "Vorwaerts(100)\n" );
	text_out_buffer->text( "" );

	// Statusanzeigen

	img_conn_off = new Fl_GIF_Image( "images/conn_off.gif" );
	img_conn_on = new Fl_GIF_Image( "images/conn_on.gif" );

	conn_state = new Fl_Box( 445, 317, 225, 40, "Verbindungsstatus" );
	conn_state->image( img_conn_off );

	// ...fr die Batteriespannung

	battery_voltage = new Fl_Value_Output( 566, 158, 100, 25, "Batteriespannung (in V):" ); 
	battery_voltage->step( 0 );
	battery_voltage->minimum( 0 );
	battery_voltage->maximum( 8.6 );
	battery_voltage->textcolor( FL_INACTIVE_COLOR );

	// ...fr die Sensoren

	robot_sigma_x = new Fl_Value_Output( 566, 190, 100, 25, "Fehler x-Richtung:" );
	robot_sigma_x->step( 0 );

	robot_sigma_y = new Fl_Value_Output( 566, 220, 100, 25, "Fehler y-Richtung:" );
	robot_sigma_y->step( 0 );

	robot_sigma_angle = new Fl_Value_Output( 566, 250, 100, 25, "Fehler Winkel:" );
	robot_sigma_angle->step( 0 );


	// Moduladressen holen

	mod_log = dynamic_cast<module_log*>( module_master::ptr().get_module( "module_log" ) );
	mod_net = dynamic_cast<module_net*>( module_master::ptr().get_module( "module_net" ) );
	mod_conf = dynamic_cast<module_conf*>( module_master::ptr().get_module( "module_conf" ) );
	mod_parser = dynamic_cast<module_parser*>( module_master::ptr().get_module( "module_parser" ) );
	mod_map = dynamic_cast<module_map*>( module_master::ptr().get_module( "module_map" ) );

	// Buttons

	btn_send = new Fl_Button( 430, 38, 220, 26, "Quelltext senden" );
	btn_send->tooltip( "bertrgt den Quelltext zum Roboter" );
	btn_send->callback( cb_cmd_send_code, this );

	btn_anima = new Fl_Button( 430, 74, 220, 25, "Roboterprogramm bertragen" );
	btn_anima->tooltip( "Programmiert den Roboter neu und ldt das erstellte Programm als Firmware" );
	btn_anima->callback( cb_cmd_anima, this );

	btn_halt = new Fl_Button( 430, 110, 220, 26, "Not-Aus" );
	btn_halt->tooltip( "Lsst den Roboter alle Befehle abbrechen, die er bisher erhalten hat. Dies stoppt die Ausfhrung aller Tasks." );
	btn_halt->callback( cb_cmd_abort, this );


	// Typen fr Netzwerknachrichten auslesen
	
	unsigned int tmp = 0;
	istringstream type_conv;
	
	type_conv.str( (string)mod_conf->da_datum( "START_PHASE_1" ) );
	type_conv >> tmp;
	MSG_START_PHASE_1 = static_cast<unsigned char>( tmp );

	type_conv.clear();
	type_conv.str( (string)mod_conf->da_datum( "START_PHASE_2" ) );
	type_conv >> tmp;
	MSG_START_PHASE_2 = static_cast<unsigned char>( tmp );

	type_conv.clear();
	type_conv.str( (string)mod_conf->da_datum( "ROBOTERBEFEHL" ) );
	type_conv >> tmp;
	MSG_PREPARE = static_cast<unsigned char>( tmp );
	
	type_conv.clear();
	type_conv.str( (string)mod_conf->da_datum( "ABBRUCH" ) );
	type_conv >> tmp;
	MSG_ABORT = static_cast<unsigned char>( tmp );

	type_conv.clear();
	type_conv.str( (string)mod_conf->da_datum( "MASZGESCHNEIDERTER_BEFEHL" ) );
	type_conv >> tmp;
	MSG_CODE = static_cast<unsigned char>( tmp );

	// Die Elemente im Men ndern sich whrend des Betriebes nicht und sind
	// daher statisch definiert.

	Fl_Menu_Item menu_items[] = 
	{
		{ "&Datei", 0, NULL, NULL, FL_SUBMENU },
		{ "&Verbindung herstellen", FL_CTRL + 'o', cb_file_connect, this },
		{ "B&eenden", FL_CTRL + 'q', cb_file_quit, this }, 
		{ NULL }, // Ende des "Datei"-Mens

		{ "&Befehle", 0, NULL, NULL, FL_SUBMENU },
		{ "&Not-Aus", FL_F+1, cb_cmd_abort, this },
		{ "Roboterprogramm &senden", 0, cb_cmd_anima, this },
		{ "Roboter &reanimieren", 0, cb_cmd_reanima, this },
		{ "Roboter &abschalten", 0, cb_cmd_inanima, this, FL_MENU_DIVIDER },
		{ "&Quelltext senden", FL_F+5, cb_cmd_send_code, this },
		{ NULL }, // Ende des "Code"-Mens 

#ifdef DEBUG			
		{ "&Debug", 0, NULL, NULL, FL_SUBMENU },
		{ "&Kartentest", FL_F+2, cb_debug_map, this },
		{ "&Hindernistest", FL_F+3, cb_debug_obstacles, this },
		{ NULL }, // Ende des "Debug"-Mens 
#endif

		{ "D&iagnose", 0, NULL, NULL, FL_SUBMENU },
		{ "Karte speichern", FL_F+6, cb_diag_dump_map, this },
		{ NULL }, // Ende des "Diagnose"-Mens

		{ "&Hilfe", 0, NULL, NULL, FL_SUBMENU },
		{ "&ber", 0, cb_help_info, NULL, NULL },
		{ NULL }, // Ende des "Hilfe"-Mens
		
		{ NULL }
	};

	menu = new Fl_Menu_Bar( 0, 0, width, 30 );
	menu->copy( menu_items );

	Fl::add_timeout( 1.0, cb_check_net_msg, this );
	Fl::add_timeout( 1.0, cb_ping, this );
	Fl::add_timeout( 1.0, cb_check_errors, this );
	Fl::add_timeout( 1.0 / 25.0, cb_redraw_gl_window, this );

	waiting_for_params = false;
	waiting_for_execution = false;
}

/// Destruktor, der den Speicher wieder freigibt. Dies ist ein Muss,
/// ansonsten wird das GUI beispielsweise unter Windows nach mehreren
/// Aufrufen nicht mehr funktionieren. 
GUI_Window::~GUI_Window( void )
{
	delete gl_win;
	
	delete text_out;
	delete text_out_buffer;
	delete text_in;
	delete text_in_buffer;

	delete robot_sigma_x;
	delete robot_sigma_y;
	delete robot_sigma_angle;

	delete menu;

	delete img_conn_on;
	delete img_conn_off;
	delete conn_state;

	delete battery_voltage;
	delete btn_send;
	delete btn_anima;
	delete btn_halt;
}


/// Fordert die bentigten Module vom Module Master an und bereitet diverse
/// Einstellungen vor.
///
/// ACHTUNG: Da hier die Zeichenfaktoren in x- und y-Richtung berechnet werden,
/// muss die Karte vor Aufruf dieses Konstruktors, also vor Start des GUI,
/// einsatzbereit sein.
///
/// @param x X-Position des Fensters
/// @param y Y-Position des Fenters
/// @param width Breite des Fenters
/// @param height Hhe des Fenters
/// @param title Titel des Fenters
GUI_GL_Window::GUI_GL_Window( int x, int y, int width, int height, char* title ) 
: Fl_Gl_Window( x, y, width, height, title )
{
	map = dynamic_cast<module_map*>( module_master::ptr().get_module( "module_map" ) );
	if( !map )
		module_master::ptr().broadcast( "*GUI_GL_Window: Kein Kartenmodul vorhanden?\n" );

	mod_conf = dynamic_cast<module_conf*>( module_master::ptr().get_module( "module_conf" ) );	
	if( !mod_conf )
		module_master::ptr().broadcast( "*GUI_GL_Window: Kein Konfigurationsmodul vorhanden?\n" );

	ROBOTER_BREITE = atof( mod_conf->da_datum( "ROBOTER_BREITE" ) );
	ROBOTER_LAENGE = atof( mod_conf->da_datum( "ROBOTER_LAENGE" ) );
	ROBOTER_MITTE_X = atof( mod_conf->da_datum( "ROBOTER_MITTE_X" ) );
	ROBOTER_MITTE_Y = atof( mod_conf->da_datum( "ROBOTER_MITTE_Y" ) );

	zoom = 1.0;

	x1 = 0.0;
	x2 = width;
	y1 = 0.0;
	y2 = height;

	draw_factor_x = w() / static_cast<double>( map->get_width() );
	draw_factor_y = h() / static_cast<double>( map->get_height() );

	// Mit dem Parameter PFAD_LEBENSZEIT wird angegeben, wie lange Roboterpfade
	// auf dem Bildschirm erscheinen. Dazu wird ein Zerfallsprozess angenommen, der nach 
	// der Hlfte der Zeit bereits zu 7/8 transparent ist. Damit ist das Ausblenden
	// sehr sanft.

	double path_ttl = atoi( mod_conf->da_datum( "PFAD_LEBENSDAUER" ) );
	if( path_ttl == 0.0 )
		path_ttl = 1.0;

	decay_e = exp( 1.0 );
	decay_lambda = ( log( 8.0 ) / log( decay_e ) ) / ( ( path_ttl / 2.0 ) * 1000.0 );
}

/// Ldt eine RAW-Datei von der Festplatte, die als Textur fr OpenGL-Objekte
/// verwendet werden kann. Momentan nicht verwendet.
///
/// @param filename Pfad zur Datei
/// @param width Breite der Texturdatei
/// @param height Hhe der Texturdatei
/// @return Eine Textur, die von OpenGL-Funktionen verwendet kann. Wenn 
/// 0 zurckgegeben wird, so ist ein Fehler aufgetreten.
unsigned int GUI_GL_Window::load_raw_texture( const char* filename, int width, int height )
{
	unsigned int texture = 0;
    	unsigned char * data;
	FILE * file;

    	file = fopen( filename, "rb" );
    	if ( file == NULL )
		return 0;

	// Texturdaten lesen...
    	
	data = new unsigned char[ width * height * 3 ];
	
	fread( data, width * height * 3, 1, file );
	fclose( file );

	glGenTextures( 1, &texture );
	glBindTexture( GL_TEXTURE_2D, texture );

	// Keine Mipmaps. Das wre nicht plattformunabhngig...
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );

	glTexImage2D( GL_TEXTURE_2D, 3, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data );

	delete[] data;
	return (texture);
}

/// Setzt den Zoomfaktor der Karte neu. Negative Werte sind
/// unzulssig.
///
/// @param new_zoom Der neue Zoomfaktor (> 0.0 )
///
/// @return Der neue Zoomfaktor oder -1.0 im Fehlerfall.
float GUI_GL_Window::set_zoom( float new_zoom )
{
	if( new_zoom <= 0.0 )
		return( -1.0 );
	
	zoom = new_zoom;
	return( zoom );	
}

/// Gibt den aktuellen Zoomfaktor aus.
///
/// @return Der Zoomfaktor als float.
float GUI_GL_Window::get_zoom( void )
{
	return( zoom );
}

/// Setzt die Position des Kartenauschnitts. Die Parameter
/// beschreiben die Eckpunkte des Rechtecks, in dem die Karte
/// bzw. das Modell dargestellt wird.
///
/// @param x1 x-Koordinate der linken unteren Ecke
/// @param x2 x-Koordinate der rechten oberen Ecke
/// @param y1 y-Koordinate der linken unteren Ecke
/// @param y2 y-Koordinate der rechten oberen Ecke
void GUI_GL_Window::set_view( double x1, double x2, double y1, double y2 )
{
	this->x1 = x1;
	this->x2 = x2;
	this->y1 = y1;
	this->y2 = y2;
}

/// Setzt den Kartenauschnitt auf die Standarddarstellung zurck. Diese stellt
/// die ganze Karte dar. Der Zoomfaktor wird durch diese Funktion ebenso
/// auf 1.0 zurckgesetzt.
void GUI_GL_Window::reset_view( void )
{
	zoom = 1.0;
	x1 = 0.0;
	x2 = w();
	y1 = 0.0;
	y2 = h();
}

/// Liefert die Differenz zweier Zeitobjekte in Millisekunden. Die Funktion difftime aus der
/// Standardbibliothek kann nur eine Differenz in Sekunden berechnen.
///
/// @param time1 Erstes Ergebnis der Zeitmessung
/// @param time2 Zweites Ergebnis der Zeitmessung
///
/// @return Die Zeitdifferenz in Millisekunden als unsigned long
unsigned long GUI_GL_Window::difftime_msec( timeval time1, timeval time2 )
{
	return(   time2.tv_sec * 1000 + time2.tv_usec / 1000
		- time1.tv_sec * 1000 - time1.tv_usec / 1000 );
}

/// Zeichnet die Karte sowie den Roboter. Die Darstellung kann in spteren
/// Versionen ohne Probleme um eine dritte Dimension ergnzt werden.
void GUI_GL_Window::draw( void )
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	unsigned int type = 0;

	// Schaltet auf den Projektionsmodus um und zeichnet die Karte.

	glMatrixMode(GL_PROJECTION);
    	glOrtho( x1 / zoom, x2 / zoom, y1 / zoom, y2 / zoom, 0.0, 1.0);

	// Hier werden zunchst alle Hindernisse gezeichnet
	
	glColor3f( 1.0, 1.0, 1.0 );

	double x1, y1; // Startpunkt der Strecke
	double x2, y2; // Endpunkt der Strecke

	double x3, y3; // Linie in Auslsungsrichtung

	double _sin;
	double _cos;

	double dir_x;
	double dir_y;

	if( map )
	{
		vector<map_info> map_data = map->get_map_data();

		glLineWidth( 1.0 );
		glEnable( GL_LINE_SMOOTH );
		glBegin( GL_LINES );
		for( size_t i = 0; i < map_data.size(); i++ )
		{
			// zeichnet eine groe Linie in Richtung des Hindernis

			x1 = -sin( map_data[ i ].angle * PI / 180.0 ) * 0.5 * map_data[ i ].width + map_data[ i ].x;
			y1 = cos( map_data[ i ].angle * PI / 180.0 ) * 0.5 * map_data[ i ].width + map_data[ i ].y;
			
			x2 = sin( map_data[ i ].angle * PI / 180.0 ) * 0.5 * map_data[ i ].width + map_data[ i ].x;
			y2 = -cos( map_data[ i ].angle * PI / 180.0 ) * 0.5 * map_data[ i ].width + map_data[ i ].y;
 
			glVertex3f( x1 * draw_factor_x, y1 * draw_factor_y, 0.0 );
			glVertex3f( x2 * draw_factor_x, y2 * draw_factor_y, 0.0 );

			// zeichnet eine kleine Linie in Richtung des Hindernis
			
			glVertex3f( 	map_data[ i ].x * draw_factor_x,
					map_data[ i ].y * draw_factor_y,
					0.0 );
			
			x3 = cos( map_data[ i ].angle * PI / 180.0 ) * 0.25 * map_data[ i ].width + map_data[ i ].x;
			y3 = sin( map_data[ i ].angle * PI / 180.0 ) * 0.25 * map_data[ i ].width + map_data[ i ].y;
			
			glVertex3f( x3 * draw_factor_x, y3 * draw_factor_y, 0.0 );
		}
		
		glEnd();
		glDisable( GL_LINE_SMOOTH );
	}

	// Den Roboter wollen wir auch noch zeichnen. Hier knnte man sich
	// wohl noch etwas Spektakulreres einfallen lassen.

	glColor3f( 1.0, 1.0, 0.0 );
	robot_data robot = map->get_robot_data();
	
	// Vektoren fr die einzelnen Punkte des Rechtecks definieren

	x1 = - ROBOTER_MITTE_X;
	y1 = - ROBOTER_MITTE_Y;

	x2 =   ROBOTER_BREITE - ROBOTER_MITTE_X;
	y2 = - ROBOTER_MITTE_Y;

	x3 = - ROBOTER_MITTE_X;
	y3 =   ROBOTER_LAENGE - ROBOTER_MITTE_Y;

	double x4 =   ROBOTER_BREITE - ROBOTER_MITTE_X;
	double y4 =   ROBOTER_LAENGE - ROBOTER_MITTE_Y;

	_sin = sin( robot.angle * PI / 180.0 );
	_cos = cos( robot.angle * PI / 180.0 );

	double nx1 = x1 * _cos - y1 * _sin; 
	double ny1 = x1 * _sin + y1 * _cos;

	double nx2 = x2 * _cos - y2 * _sin;
	double ny2 = x2 * _sin + y2 * _cos;

	double nx3 = x3 * _cos - y3 * _sin;
	double ny3 = x3 * _sin + y3 * _cos;

	double nx4 = x4 * _cos - y4 * _sin;
	double ny4 = x4 * _sin + y4 * _cos;

	glBegin( GL_QUADS );
	glColor3f( 1.0, 0.0, 0.0 );
	glVertex3f( draw_factor_x * ( nx3 + robot.x ), draw_factor_y * ( ny3 + robot.y ), 0.0 );
	glVertex3f( draw_factor_x * ( nx4 + robot.x ), draw_factor_y * ( ny4 + robot.y ), 0.0 );
	glColor3f( 0.0, 0.0, 1.0 );
	glVertex3f( draw_factor_x * ( nx2 + robot.x ), draw_factor_y * ( ny2 + robot.y ), 0.0 );
	glVertex3f( draw_factor_x * ( nx1 + robot.x ), draw_factor_y * ( ny1 + robot.y ), 0.0 );
	glEnd();

	/*
	SPIELEREI: Zeichnet den Namen des Roboters

	glDisable(GL_DEPTH_TEST);
        const char *p;
        gl_font(1, 12);
        glColor3f(1.0, 1.0, 1.0);
        glRasterPos3f( robot.x * draw_factor_x, robot.y * draw_factor_y, 0.0);   p = robot.name.c_str();   gl_draw(p, strlen(p));
	*/

	// Vorbereitung zum Zeichnen der Pfade. Es gibt wahrscheinlich noch bessere Mglichkeiten,
	// alle Pfade abzuspeichern, aber dies fhrt hier zu weit. Denkbar wren beispielsweise
	// Binre Suchbume...
	
	vector<robot_path> paths = map->get_paths();

	// Beseitigt den "Fog of War" auf den Wegen, die der Roboter bereits
	// gefahren ist.
		
	glColor3f( 0.0, 1.0, 0.0 );

	double v1, v2;
	double vector_length;

	for( size_t i = 0; i < paths.size(); i++ )
	{
		if( paths[ i ].get_permanent() )
		{
			v1 = paths[ i ].get_source_y() - paths[ i ].get_target_y();
			v2 = paths[ i ].get_target_x() - paths[ i ].get_source_x();

			vector_length = sqrt( v1 * v1 + v2 * v2 );
			v1 /= vector_length;
			v2 /= vector_length;
			
			x1 = paths[ i ].get_source_x() + v1 * ROBOTER_BREITE * 0.5;
			y1 = paths[ i ].get_source_y() + v2 * ROBOTER_BREITE * 0.5;

			x2 = paths[ i ].get_source_x() - v1 * ROBOTER_BREITE * 0.5;
			y2 = paths[ i ].get_source_y() - v2 * ROBOTER_BREITE * 0.5;
			
			x3 = paths[ i ].get_target_x() - v1 * ROBOTER_BREITE * 0.5;
			y3 = paths[ i ].get_target_y() - v2 * ROBOTER_BREITE * 0.5;

			x4 = paths[ i ].get_target_x() + v1 * ROBOTER_BREITE * 0.5;
			y4 = paths[ i ].get_target_y() + v2 * ROBOTER_BREITE * 0.5;


			glBegin( GL_QUADS );
			glVertex3f( x1 * draw_factor_x, y1 * draw_factor_y, -0.25 );
			glVertex3f( x2 * draw_factor_x, y2 * draw_factor_y, -0.25 );
			glVertex3f( x3 * draw_factor_x, y3 * draw_factor_y, -0.25 );
			glVertex3f( x4 * draw_factor_x, y4 * draw_factor_y, -0.25 );
			glEnd();
		 }
	}

	// Zeichnet die Verbindungslinien, so es welche gibt. Diese werden
	// nach einer zuvor eingestellten Zeit wieder gelscht.
	
	glLineWidth( 5.0 );
	glEnable (GL_LINE_SMOOTH);

	glEnable( GL_BLEND );
	glDisable( GL_DEPTH_TEST );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	timeval now;
	gettimeofday( &now, NULL );

	for( size_t i = 0; i < paths.size(); i++ )
	{
		if( paths[ i ].get_status() && !paths[ i ].get_permanent() )
		{
			glColor4f( 1.0, 0.0, 0.0, 1.0 * pow( decay_e, -decay_lambda * difftime_msec( paths[ i ].get_creation_time(), now ) ) );
			
			glBegin( GL_LINES );
			glVertex3f( paths[ i ].get_source_x() * draw_factor_x, paths[ i ].get_source_y() * draw_factor_y, 0.0 );
			glVertex3f( paths[ i ].get_target_x() * draw_factor_x, paths[ i ].get_target_y() * draw_factor_y, 0.0 );
			glEnd();
		 }
	}

	glEnable( GL_DEPTH_TEST );
	glDisable( GL_BLEND );
	glDisable( GL_LINE_SMOOTH );
}

/// Callback-Funktion, die sich auf den Menpunkt Hilfe->ber bezieht
/// und ein kurzes Fenster mit Informationen zum Programm anzeigt. Mehr
/// eine Spielerei und ein Test fr das GUI.
///
/// @param w Das Widget, das den Callback verursachte.
/// @param dummy Nicht verwendet.
void GUI_Window::cb_help_info( Fl_Widget* w, void* dummy )
{
	fl_message(	"Eukalyptos Version 1.0\n\n"
			"Dieses Programm dient zur Netzwerkfernsteuerung von\n"
			"Lego-RCX-2-Robotern und wurde von Lutz Bch, Bastian Rieck und\n"
			"Carlos Franke im Rahmen eines Praktikums am Interdisziplinren\n"
			"Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt\n"
			"Heidelberg geschaffen." );
}

/// Callback-Funktion, die das Fenster schliet. Damit wird dann sofort das
/// Programm beendet.
///
/// @param w Das Widget, das den Callback ausgelst hat. Nicht verwendet.
/// @param ptr Ein Zeiger auf das GUI. Die Funktion ruft dann die callback-Funktion
/// des Hauptfensters auf.
void GUI_Window::cb_file_quit( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	gui->do_callback();
}

/// Callback-Funktion, die sich auf den Menpunkt Datei->Verbindung herstellen
/// bezieht. Sie zeigt ein Fenster an, das den Benutzer auffordert, eine Adresse
/// der Art "ServerName:Port" einzugeben. Danach versucht das GUI, eine Verbindung
/// ber das Netzwerkmodul herzustellen.
///
/// @param w Das Widget, das den Callback ausgelst hat. Nicht verwendet.
/// @param ptr Ein Zeiger auf das GUI. Nur so ist der Zugriff auf Membervariablen
/// mglich.
void GUI_Window::cb_file_connect( Fl_Widget* w, void* ptr )
{
	const char* addr_data = fl_input( 	"Bitte geben Sie die Adresse des Servers in der Form\n"
						"\'servername:port\' ein. Sie knnen auch eine IP-Adresse\n"
						"angeben:\n",
						NULL );

	if( addr_data == NULL )
		return;

	
	// berprfen, ob die Adresse syntaktisch gltig ist
	if( strstr( addr_data, ":" ) == NULL )
	{
		fl_alert( "Dieses Eingabeformat vermag ich nicht zu akzeptieren." );
		return;
	}

	// Port und Server extrahieren.

	char* server_addr = new char[ strlen( addr_data ) - strlen( strstr( addr_data, ":" ) ) + 1 ];
	for( unsigned int i = 0; i < strlen( addr_data ); i++ )
	{
		if( addr_data[ i ] != ':' )
			server_addr[ i ] = addr_data[ i ];
		else
		{
			server_addr[ i ] = '\0';
			break;	
		}
	}
	
	unsigned int port = atoi( strstr( addr_data, ":" ) + 1 );

	// Einfach mal die Verbindung herstellen. Um Fehler kmmert sich
	// das Netzwerkmodul.

	if( ptr != NULL )
	{
		GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );

		module_net* mod_net = gui->mod_net; // Abkrzung
		mod_net->connect( server_addr, port );

		gui->conn_state->deimage();
		gui->conn_state->image( gui->img_conn_on );
		
		gui->conn_state->redraw();
		gui->conn_state->redraw_label();

		gui->waiting_for_params = false;
		gui->waiting_for_execution = false; 
	}
	else
		module_master::ptr().broadcast( "*GUI_Window: Kein Netzwerkmodul geladen?\n" );
		 
	delete[] server_addr;
}

/// Sendet die Nachricht zum Einschalten des Roboters. Dabei wird 
/// automatisch ein neues Programm auf den Roboter bertragen und
/// sogleich ausgefhrt.
///
/// @param w Das Widget, das den Callback verursacht hat
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_cmd_anima( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net;

	if( !mod_net->has_connection() )
	{
		module_master::ptr().broadcast( "#GUI_Window: In cb_cmd_anima: Ohne Netzwerkverbindung kann ich den Roboter nicht einschalten.\n" );
		return;
	}
	
	net_msg_hdr msg( MSG_TYPE_SYSTEM, MSG_ANIMA );
	mod_net->send_msg( reinterpret_cast<const char*>( &msg ), sizeof( net_msg_hdr ) );

	gui->waiting_for_execution = false;
}

/// Reanimiert den Roboter. Damit wird das Hauptprogramm
/// wieder ausgefhrt, sodass der Roboter nach einem Fehler
/// wieder verwendet werden kann. 
///
/// @param w Das Widget, das den Callback verursacht hat
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_cmd_reanima( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net;

	if( !mod_net->has_connection() )
	{
		module_master::ptr().broadcast( "#GUI_Window: In cb_cmd_reanima: Ohne Netzwerkverbindung kann ich den Roboter nicht reanimieren.\n" );
		return;
	}
	
	net_msg_hdr msg( MSG_TYPE_SYSTEM, MSG_REANIMA );
	mod_net->send_msg( reinterpret_cast<const char*>( &msg ), sizeof( net_msg_hdr ) );

	gui->waiting_for_execution = false;
}

/// Sendet die Nachricht zum Abschalten des Roboters.
///
/// @param w Das Widget, das den Callback verursacht hat
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_cmd_inanima( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net;
	
	if( !mod_net->has_connection() )
	{
		module_master::ptr().broadcast( "#GUI_Window: In cb_cmd_inanima: Ohne Netzwerkverbindung kann ich den Roboter nicht ausschalten.\n" );
		return;
	}

	net_msg_hdr msg( MSG_TYPE_SYSTEM, MSG_INANIMA );
	mod_net->send_msg( reinterpret_cast<const char*>( &msg ), sizeof( net_msg_hdr ) );

	gui->waiting_for_execution = false;
}

/// Bringt den Roboter dazu, alle Aktionen abzubrechen und das aufgespielte
/// Programm neuzustarten. Diese Funktion sollte nur dann genutzt werden,
/// wenn etwas in der Programmierung schiefgelaufen ist.
///
/// @param w Das Widget, das den Callback verursacht hat
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_cmd_abort( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net;
	
	if( !mod_net->has_connection() )
	{
		module_master::ptr().broadcast( "#GUI_Window: In cb_cmd_halt: Ohne Netzwerkverbindung kann ich den Roboter nicht ausschalten.\n" );
		return;
	}

	net_msg_hdr msg( MSG_TYPE_ROBOT, gui->MSG_ABORT );
	mod_net->send_msg( reinterpret_cast<const char*>( &msg ), sizeof( net_msg_hdr ) );

	gui->waiting_for_execution = false;
}


/// Liest den vom Benutzer eingegebenen Quelltext und ordnet ihm zuvor 
/// definierte Befehle zu. Diese werden an den Server-PC gesandt, der aus
/// ihnen ein auf dem Roboter ausfhrbares Programm erstellt.
///
/// Der Quelltext ist der Text, den der Benutzer in das 
/// Fl_Text_Editor-Widget namens "text_in" geschrieben hat.
/// geschrieben hat.
///
/// @param w Das Widget, das den Callback verursacht hat.
/// @param ptr Ein Zeiger auf das GUI. Wird bentigt wegen des Zugriffs
/// auf Membervariablen. 
void GUI_Window::cb_cmd_send_code( Fl_Widget* w, void* ptr )
{
	if( ptr == NULL )
		return;

	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );

	// Der Einfachheit halber, um nicht immer den GUI-Zeiger
	// verwenden zu mssen.	
	
	module_conf* mod_conf = gui->mod_conf;
	module_net* mod_net = gui->mod_net;
	module_parser* mod_parser = gui->mod_parser;

	Fl_Text_Buffer* text_in_buffer = gui->text_in_buffer;

	if( mod_conf == NULL || mod_net == NULL || text_in_buffer == NULL )
		return;

	if( !mod_net->has_connection() )
	{
		module_master::ptr().broadcast( "#GUI_Window: Keine Befehlsbertragung zum Roboter mglich, da keine Netzwerkverbindung besteht.\n" );
		
		gui->waiting_for_execution = false;
		return;
	}
	else if( gui->waiting_for_execution )
	{
		fl_alert( 	"Der Roboter fhrt gerade Befehle aus. Nur Geduld! Falls der Roboter\n"
				"nicht mehr antwortet, sollte das Programm neu aufgespielt werden.\n" );
		return;
	}

	string sourcecode = text_in_buffer->text();
	vector<saved_net_msg> res = mod_parser->parse( sourcecode );
	
	// Es ist ein Fehler beim Kompilieren aufgetreten
	if( res.size() == 0 )
	{
		module_master::ptr().broadcast( "*GUI_Window: Fehler beim Kompilieren des Benutzerquelltextes. Bitte Syntax berprfen.\n" );
		return;
	}
	
	// Zur Zeitmessung muss der Vektor in ein anderes
	// Format konvertiert werden.

	vector<saved_net_msg> saved_messages;
	saved_net_msg msg;
	for( size_t i = 0; i < res.size(); i++ )
	{
		msg.header = res[ i ].header;
		msg.params = res[ i ].params;
		saved_messages.push_back( msg );
	}

	// Die Zeit wird dann zum Server geschickt.
	
	double execution_time = gui->mod_map->estimate_execution_time( saved_messages );
	ostringstream conv;
	conv << execution_time;

	net_msg_hdr time_msg( MSG_TYPE_SYSTEM, MSG_EXECUTION_TIME, conv.str().size() );
	mod_net->send_msg( reinterpret_cast<const char*>( &time_msg ), sizeof( net_msg_hdr ) );
	mod_net->send_msg( conv.str().c_str(), conv.str().size() );

	// Nach MSG_CODE suchen

	saved_net_msg code_msg;
	code_msg.header.set_type( 0 );
	for( size_t i = 0; i < res.size(); i++ )
	{
		if( res[ i ].header.get_cmd() == gui->MSG_CODE )
		{
			code_msg.header  = res[ i ].header;
			code_msg.params = res[ i ].params;
			
			res.erase( res.begin() + i, res.begin() + i + 1 );
		}
	}

	// Nachrichten schicken, wenn nicht nur eine Schneiderleinnachricht
	// gesendet werden soll.
	//
	// Wichtig ist, dass das Flag zur Befehlsausfhrung gesetzt ist. Andernfalls
	// kann der Benutzer den Roboter und den Server durch neue Befehle aus
	// dem Tritt bringen.

	gui->waiting_for_execution = true;

	if( res.size() == 0 && code_msg.header.get_type() != 0 )
	{
		module_master::ptr().broadcast( "#GUI_Window: Versende Schneiderleinbefehle.\n" );

		mod_net->send_msg( reinterpret_cast<const char*>( &code_msg.header ), sizeof( net_msg_hdr ), MSG_NOSIGNAL );
		mod_net->send_msg( code_msg.params.c_str(), code_msg.params.size(), MSG_NOSIGNAL );
		return;
	}

	net_msg_hdr start_msg( MSG_TYPE_ROBOT, gui->MSG_START_PHASE_1 );
	mod_net->send_msg( reinterpret_cast<const char*>( &start_msg ), sizeof( net_msg_hdr ) );

	net_msg_hdr prepare_msg( MSG_TYPE_ROBOT, gui->MSG_PREPARE );
	mod_net->send_msg( reinterpret_cast<const char*>( &prepare_msg ), sizeof( net_msg_hdr ) );

	gui->buffered_messages.clear();
	saved_net_msg save_msg;
	
	for( size_t i = 0; i < res.size(); i++ )
	{
		mod_net->send_msg( 	reinterpret_cast<const char*>( &res[ i ].header ), 
					sizeof( net_msg_hdr ) );
		mod_net->send_msg( res[ i ].params.c_str(), res[ i ].params.size() );

		// Die Nachricht im Puffer speichern, sodass das Modell spter
		// berprfen kann, welche Befehle entgegegenommen wurden und wie
		// das Ausfhrungsergebnis derselbigen war.

		save_msg.header = res[ i ].header;
		
		if( res[ i ].header.get_size() > 0 )
			save_msg.params = res[ i ].params;
		else
			save_msg.params = "";	
		
		gui->buffered_messages.push_back( save_msg );
	}

	// Diese Nachricht bringt den Roboter dazu, alle Befehle
	// umgehend auszufhren

	start_msg.set_cmd( gui->MSG_START_PHASE_2 );	
	mod_net->send_msg( reinterpret_cast<const char*>( &start_msg ), sizeof( net_msg_hdr ) );

	// Gab es Schneiderleinbefehle?
	if( code_msg.header.get_type() != 0 )
	{
		mod_net->send_msg( reinterpret_cast<const char*>( &code_msg.header ), sizeof( net_msg_hdr ), MSG_NOSIGNAL );
		mod_net->send_msg( code_msg.params.c_str(), code_msg.params.size(), MSG_NOSIGNAL );
	}
}

/// berprft in gewissen Abstnden, ob neue Nachrichten beim Netzwerkmodul
/// eingetroffen sind.
///
/// @param ptr Ein Zeiger auf das GUI.
void GUI_Window::cb_check_net_msg( void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net; // Abkrzung...
	
	Fl::repeat_timeout(1.0, cb_check_net_msg, gui );
	if( !mod_net->has_connection() )
		return;
	
	net_msg_hdr hdr;
	char* params = NULL;
	string param_temp; // Manchmal ist es einfacher, mit STL-Strings zu arbeiten

	robot_data robot;
	double battery_voltage = 0.0;
	
	int bytes = mod_net->has_msg();

	// Hier ndert sich eigentlich nur die Anzahl der erwarteten Bytes.
	if( gui->waiting_for_params )
	{
		if( bytes >= gui->old_hdr.get_size() )
		{
			// Parameter empfangen. Das Paket wird weiter unten im
			// Code tatschlich behandelt.

			params = new char[ gui->old_hdr.get_size() + 1 ];
			memset( params, 0, gui->old_hdr.get_size() + 1 );
			mod_net->recv_msg( params, gui->old_hdr.get_size() );

			hdr = gui->old_hdr;
			gui->waiting_for_params = false;
		}
	}
	else
	{
		if( bytes >= sizeof( net_msg_hdr ) )
		{
			mod_net->recv_msg( reinterpret_cast<char*>( &hdr ), sizeof( hdr ) );

#ifdef DEBUG
			fprintf( stderr, "GUI_Window: Nachricht erhalten:\n"
					 "\t(*) Typ = %i\n"
					 "\t(*) Befehl = %i\n",
			 		 hdr.get_type(),
					 hdr.get_cmd() );
#endif			
				
			// Wenn der Nachricht Parameter beigefgt sind, so warten wir
			// auf diese.
			if( hdr.get_size() > 0 )
			{
				gui->waiting_for_params = true;
				gui->old_hdr = hdr;
			}
		}
	}

	// Jetzt kann das Paket behandelt werden.
	if( !gui->waiting_for_params )
	{
		switch( hdr.get_cmd() )
		{
			// Diese Nachricht enthlt das Datalog mit allen Befehlen, die vom Roboter
			// ausgefhrt wurden. Im Fehlerfall ist dort die Zeit der Sensorauslsung
			// zu finden.

			case MSG_EXECUTED_COMMANDS:
				
				fprintf( stderr, "GUI: parse_datalog beginnt nun.\n" );

				gui->mod_map->parse_datalog( gui->buffered_messages, params );
				gui->battery_voltage->value( gui->mod_map->get_battery_voltage() / 1000.0 );

				gui->waiting_for_execution = false;

				// Fehler ausgeben
				
				robot = gui->mod_map->get_robot_data();
				gui->robot_sigma_x->value( robot.sigma_x );
				gui->robot_sigma_y->value( robot.sigma_y );
				gui->robot_sigma_angle->value( robot.sigma_angle ); 

				break;

			// Diese Nachricht enthlt die aktuelle Batteriespannung des Roboters in
			// Volt.	
			case MSG_BATTERY_VOLTAGE:
				gui->battery_voltage->value( atof( params ) / 1000.0 );
				break;
				
			// Lognachrichten werden ans Logmodul weitergereicht
			case MSG_LOG:

				// Eine Sonderbehandlung fr die Nachricht, die die Batteriespannung
				// enthlt, fr die das Robotermodul kompiliert wurde. Diese wird
				// nicht an das Ausgabemodul weitergereicht.

				param_temp = params;
				if( param_temp.find( "Batteriespannung:" ) != string::npos )
				{
					param_temp.erase( 0, param_temp.find( "Batteriespannung:" ) + strlen( "Batteriespannung:" ) + 1 );
					battery_voltage = atof( param_temp.c_str() );
					battery_voltage /= 1000.0;

					gui->mod_map->set_battery_voltage( battery_voltage, true );
					gui->battery_voltage->value( battery_voltage );
					break;
				}

				gui->mod_log->handle_msg( params );
				break;
					
			default:
				break;	 
		}
	}

	if( params )
	{
		delete[] params;
		params = NULL;
	}

	// Je nachdem, wie die Batteriespannung ist, wird hier das Widget, das
	// diese anzeigt, mit einer anderen Textfarbe versehen.

	if( gui->battery_voltage->value() >= 8.0 )
		gui->battery_voltage->textcolor( FL_DARK_GREEN );
	else if( gui->battery_voltage->value() >= 7.6 )
		gui->battery_voltage->textcolor( FL_DARK_YELLOW );
	else
		gui->battery_voltage->textcolor( FL_DARK_RED );		
}

/// Sendet eine Ping-Nachricht an den Server, um zu sehen, ob dieser
/// noch aktiv ist. Falls nicht, so wird die Verbindung geschlossen.
///
/// @param ptr Ein Zeiger auf das GUI. Nur so ist der Zugriff auf Membervariablen
/// mglich.
void GUI_Window::cb_ping( void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	module_net* mod_net = gui->mod_net; // Abkrzung :-)
	
	Fl::repeat_timeout( 1.0, cb_ping, gui );
	
	if( !mod_net->has_connection() )
	{
		gui->waiting_for_execution = false;
		gui->waiting_for_params = false;
		return;
	}

	net_msg_hdr hdr( MSG_TYPE_SYSTEM, MSG_PING );
	if( mod_net->send_msg( reinterpret_cast<char*>( &hdr ), sizeof( hdr ), MSG_NOSIGNAL ) < 0 )
	{
		// Wenn nicht schon getan: Status der Verbindung fr den User
		// kenntlich machen.
		if( gui->conn_state->image() != gui->img_conn_off )
		{
			gui->conn_state->deimage();
			gui->conn_state->image( gui->img_conn_off );

			gui->conn_state->redraw();
			gui->conn_state->redraw_label();
		}

		module_master::ptr().broadcast( "*GUI_Window: Keine Netzwerkverbindung verfgbar.\n" );
		mod_net->disconnect();

		gui->waiting_for_params = false;
		gui->waiting_for_execution = false;
	}
}

/// Liest eventuell angefallene Fehler, Warnungen und Informationsmeldungen vom
/// Log-Modul aus und trgt diese in Widget des GUI ein.
///
/// @param ptr Ein Zeiger auf das GUI, sodass auf alle Funktionen zugegriffen
/// werden kann.
void GUI_Window::cb_check_errors( void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	
	string msg;
	string output = gui->text_out_buffer->text();
	while( ( msg = gui->mod_log->get_oldest_msg() ) != "" )
	{
		// Es werden fr jede Nachricht "\n" angehngt, sofern diese
		// nicht bereits vorhanden sind.
		output += msg;
		if( msg.find( "\n" ) == string::npos )
			output += "\n";
	}

	gui->text_out_buffer->text( output.c_str() );
	gui->text_out->scroll( gui->text_out_buffer->count_lines( 0, gui->text_out_buffer->length() ), 0 );
	
	Fl::repeat_timeout( 1.0, cb_check_errors, ptr );
}

/// Zeichnet das OpenGL-Fenster neu. Somit knnen Animationen in
/// annehmbaren Tempo abgespielt werden. Dieser Callback wird als
/// Timeout aufgerufen, sodass in etwa 25 FPS mglich sind. Ebenso
/// ruft dieser Callback die Update-Funktion der Karte auf, sodass
/// eventuell inaktive Pfade gefunden werden.
///
/// @param ptr Ein Zeiger auf das GUI-Fenster 
void GUI_Window::cb_redraw_gl_window( void* ptr )
{
	if( ptr == NULL )
	{
		module_master::ptr().broadcast( "*GUI_Window: In cb_redraw_gl_window: Kein OpenGL verfgbar?\n" );
		return;
	}

	GUI_Window* window = reinterpret_cast<GUI_Window*>( ptr );
	window->gl_win->redraw();
	window->mod_map->update();
	
	Fl::repeat_timeout( 1.0 / 25.0, cb_redraw_gl_window, ptr );	
}

/// DEBUG:
/// Fhrt einen Kartentest durch. Diese Funktion kann nur ber das
/// Men aufgerufen werden, wenn DEBUG definiert ist.
///
/// Zum Testen der Karte werden Pfade zuflligen Typs an zuflligen
/// Stellen eingefgt. Diese werden danach zur spteren Analyse
/// in die Datei paths.dat geschrieben.
///
/// @param w Das Widget, das den Callback verursacht hat.
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_debug_map( Fl_Widget* w, void* ptr )
{
	bool permanent = static_cast<bool>( rand() % 2 );

	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	robot_path path( 	rand() % static_cast<unsigned int>( gui->mod_map->get_width() ),
				rand() % static_cast<unsigned int>( gui->mod_map->get_height() ),
				0.0,
				rand() % static_cast<unsigned int>( gui->mod_map->get_width() ),
				rand() % static_cast<unsigned int>( gui->mod_map->get_height() ),
				0.0,
				permanent );	
	
	gui->mod_map->add_path( path );
	gui->mod_map->dump_paths( "paths.dat" );
}

/// DEBUG:
/// Fhrt einen Hindernistest durch. Diese Funktion kann nur ber das
/// Men aufgerufen werden, wenn DEBUG definiert ist.
///
/// Zum Testen der Hindernisdarstellung werden Hindernisse an zuflligen
/// Stellen mit zuflligen Winkeln eingefgt.
///
/// @param w Das Widget, das den Callback verursacht hat.
/// @param ptr Ein Zeiger auf das GUI
void GUI_Window::cb_debug_obstacles( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	
	map_info obstacle;
	obstacle.x = rand() % static_cast<unsigned int>( gui->mod_map->get_width() );
	obstacle.y = rand() % static_cast<unsigned int>( gui->mod_map->get_height() );
	obstacle.width = 18.0;
	
	double sensor_angle = 90.0;
	obstacle.angle = (double)fmod( ( sensor_angle + (double)fmod( (double)rand(), 360.0 ) ), 360.0 );

	gui->mod_map->add_map_element( obstacle );
}


/// Speichert alle bisher gefahrenen Pfade des Roboters in einer Datei 
/// ab, sodass sich beispielsweise mittels Gnuplot ein genaues Profil der
/// Roboterbewegungen darstellen lsst.
///
/// @param w Widget, das den Callback ausgelst hat.
/// @param ptr Zeiger auf das GUI
void GUI_Window::cb_diag_dump_map( Fl_Widget* w, void* ptr )
{
	GUI_Window* gui = reinterpret_cast<GUI_Window*>( ptr );
	const char* filename = fl_input( "Bitte Dateinamen der Ausgabedatei eingeben: ", "paths.dat" );
	
	if( filename == NULL )
		return;
	
	gui->mod_map->dump_paths( filename );
}
