/*

	~~~~~~~~~
	~ gui.h ~
	~~~~~~~~~

Funktionen fr das GUI. Das GUI ist _kein_ Modul, sondern ein unabhngiger
Programmteil, der alle Module in einem groen Ganzen verwendet. Alle
Zeichenoperationen wurden mit FLTK und/oder OpenGL durchgefhrt.

Das GUI wurde erfolgreich kompiliert und getestet unter:
- Gentoo Linux
- FreeBSD

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/

#ifndef GUI_FLTK_H
#define GUI_FLTK_H

#include <FL/Fl.h>
#include <FL/Fl_Window.h>
#include <FL/Fl_Gl_Window.h>
#include <FL/Fl_Multiline_Input.h>
#include <FL/Fl_Text_Editor.h>
#include <FL/Fl_Text_Display.h>
#include <FL/Fl_Menu.h>
#include <FL/Fl_Menu_Bar.h>
#include <FL/Fl_Menu_Item.h>
#include <FL/Fl_PNG_Image.h>
#include <FL/Fl_GIF_Image.h>
#include <FL/Fl_Box.h>
#include <FL/Fl_Button.h>
#include <FL/Fl_Value_Output.h>

#include "module_map.h"
#include "module_log.h"
#include "module_net.h"
#include "module_conf_ConfigFile.h"
#include "module_parser.h"

/// Dieses Fenster dient dem Rendering des Modells, also der Darstellung
/// der Karte sowie der Darstellung des Roboters.
class GUI_GL_Window : public Fl_Gl_Window
{
	public:
		GUI_GL_Window( int x, int y, int width, int height, char* title );

		float set_zoom( float new_zoom );
		float get_zoom( void );

		void set_view( double x1, double x2, double y1, double y2 );
		void reset_view( void );

	private:
		double ROBOTER_BREITE;
		double ROBOTER_LAENGE;
		double ROBOTER_MITTE_Y;
		double ROBOTER_MITTE_X;
		
		float zoom;

		double x1; ///< x-Koordinate der linken unteren Ecke der Karte
		double x2; ///< x-Koordinate der rechten oberen Ecke der Karte
		double y1; ///< y-Koordinate der linken unteren Ecke der Karte
		double y2; ///<	y-Koordinate der rechten oberen Ecke der Karte

		unsigned long difftime_msec( timeval time1, timeval time2 );
		unsigned int load_raw_texture( const char* filename, int width, int height );
		
		module_map* map; ///< Die Karte mit allen Terraindaten
		module_conf* mod_conf;

		double draw_factor_x;	///< Faktor, mit dem die Karte in X-Richtung skaliert wird
		double draw_factor_y;	///< Faktor, mit dem die Karte in Y-Richtung skaliert wird
		double decay_e;		///< Speichert die Euler'sche Zahl. Diese wird verwendet, um die gespeicherten
					///< Pfade der Karte sanft auszublenden.
		double decay_lambda;	///< Gibt den Zerfalls-/ bzw. Ausblendefaktor an.
		
		void draw( void );
};

/// Dieses Fenster ist das Hauptfenster, in dem alle anderen Komponenten
/// wie beispielsweise das Modell und die Steuerelemente untergebracht sind.
class GUI_Window : public Fl_Window
{
	public:
		GUI_Window(int width, int height, char* title);
		virtual ~GUI_Window( void );

	private:
	
		GUI_GL_Window* gl_win; 			///< Fenster fr das Modell

		Fl_Text_Display* text_out;		///< Widget fr die Textausgabe
		Fl_Text_Buffer* text_out_buffer;	///< Puffer fr die Textausgabe
		Fl_Text_Editor* text_in;		///< Widget fr die Texteingabe
		Fl_Text_Buffer* text_in_buffer;		///< Puffer fr die Texteingabe
		
		Fl_Menu_Bar* menu;			///< Men
		
		Fl_GIF_Image* img_conn_on;		///< Bild, das eine aktive Netzwerkverbindung kennzeichnet
		Fl_GIF_Image* img_conn_off;		///< Bild, das eine inaktive Netzwerkverbindung kennzeichnet
		Fl_Box* conn_state;			///< Box, die den aktuellen Status der Netzwerkverbindung anzeigt

		Fl_Button* btn_send;
		Fl_Button* btn_anima;
		Fl_Button* btn_halt;
		
		Fl_Value_Output* battery_voltage;
		Fl_Value_Output* robot_sigma_x;
		Fl_Value_Output* robot_sigma_y;
		Fl_Value_Output* robot_sigma_z;
		Fl_Value_Output* robot_sigma_angle;

		module_net* mod_net;
		module_conf* mod_conf;
		module_parser* mod_parser;
		module_log* mod_log;
		module_map* mod_map;

		// Codes fr die Netzwerknachrichten
		unsigned char MSG_START_PHASE_1;
		unsigned char MSG_START_PHASE_2;
		unsigned char MSG_PREPARE;
		unsigned char MSG_CODE;
		unsigned char MSG_ABORT;

		net_msg_hdr old_hdr;		///< Dieser Header speichert einen Header mit Parametern zwischen, sodass
						///< der Callback cb_check_net_msg diese zu einem spteren Zeitpunkt lesen
						///< kann.
		bool waiting_for_params;	///< Dieses Flag wird gesetzt, wenn der Callback cb_check_net_msg noch
						///< auf die vollstndige bertragung von Parametern wartet.
		bool waiting_for_execution;	///< Dieses Flag wird gesetzt, wenn das GUI noch auf die Ausfhrung der Befehle
						///< wartet, die an den Roboter gesendet wurden. Whrend dieser Wartezeit ist es
						///< nicht mglich, andere Befehle an den Roboter zu bertragen.				
		
		/// Enthlt alle Nachrichten mit Parametern, die zu einer Konversation mit dem
		/// Roboter gehren.				
		vector<saved_net_msg> buffered_messages;			

		// Ab hier kommen die sogenannten Callback-Funktionen. Sie werden aufgerufen, wenn
		// der Benutzer eine Aktion mit bestimmten Widgets durchfhrt. Ebenso sind hier die
		// Funktionen zu finden, die in regelmigen Abstnden ausgefhrt werden.
		// Aus FLTK-internen Grnden mssen diese Funktionen allesamt statisch sein. 
		//
		// Die Nomenklatur ist wie folgt:
		// (i)  cb_(Name des Mens auf Englisch)_(Menpunkt auf Englisch)_...
		// (ii) cb_(Name des Widgets)
		//
		// Eine Callbackfunktion, die sich auf Datei->ffnen bezieht, wrde also cb_file_open
		// heien.
		//
		// Mehr Informationen zu diesen Callbacks finden sich in gui.cpp.
		
		static void cb_file_quit( Fl_Widget*, void* );
		static void cb_file_connect( Fl_Widget*, void* );
		static void cb_help_info( Fl_Widget*, void* );
		static void cb_cmd_anima( Fl_Widget*, void* ); 
		static void cb_cmd_reanima( Fl_Widget*, void* ); 
		static void cb_cmd_inanima( Fl_Widget*, void* ); 
		static void cb_cmd_abort( Fl_Widget*, void* ); 
		static void cb_cmd_send_code( Fl_Widget*, void* ); 
		static void cb_diag_dump_map( Fl_Widget*, void* );
		
		static void cb_debug_map( Fl_Widget*, void* );
		static void cb_debug_obstacles( Fl_Widget*, void* );

		static void cb_check_net_msg( void* );
		static void cb_ping( void* );
		static void cb_redraw_gl_window( void* );
		static void cb_check_errors( void* );
};

#endif
