/*
	~~~~~~~~~~~~~~
	~ module.cpp ~
	~~~~~~~~~~~~~~

Funktionen zur Modulverwaltung. Ebenso finden sich hier die Grundfunktionen der
Modulklasse.

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/

#include "module.h"

#include <stdio.h>
#include <string.h>

/// Erstellt die einzige Instanz des Singletons.
module_master::module_master( void )
{
}


/// Liefert eine Instanz des Singletons. Wenn der Singleton auf diese
/// Weise (ber eine statische Instanz) erzeugt wird, hat dies den
/// Vorteil, dass man Konstruktoren/Destruktoren verwenden kann
/// und sich nicht um Speicherverwaltung kmmern muss.
module_master& module_master::ptr( void )
{
	static module_master instance;
	return( instance );
}

/// Registriert ein Modul, indem dieses zur Datenbank hinzugefgt wird. Die
/// Anzahl der Module kann beliebig gro sein, sofern das RAM mitmacht. Beim
/// Registrieren wird das Modul durch einen Aufruf von mod_load() gleich
/// geladen.
///
/// @param m Das zu registrierende Modul.
///
/// @return 0, wenn das Modul registriert werden konnte. -1, wenn ein Fehler auftrat.
int module_master::register_module( module* m )
{
	if( m == NULL )
		return( -1 );
	
	module_info mod_info;
	strncpy( mod_info.name, m->get_name(), MODULE_NAME_LENGTH );
	mod_info.ptr = m;
	mod_info.id = 0;

	db.push_back( mod_info );
	m->mod_load();

	return( 0 );
}

/// Entfernt die Regstrierung eines Moduls aus der Datenbank. Das Modul
/// wird dabei auch gleich beendet, sodass z.B. benutzter Speicher
/// wieder freigegeben wird.
///
/// @param m Das zu entfernende Modul
///
/// @return 0, wenn das Modul erfolgreich entfernt werden konnte. -1 sonst.
int module_master::unregister_module( module* m )
{
	// Suchen...
	for( int i = 0; i < db.size(); i++ )
	 {
		 if( 	!strncmp( db[ i ].name, m->get_name(), MODULE_NAME_LENGTH ) /* Wird momentan nicht ausgewertet &&
			db[ i ].id == m->get_id() */ )
		  {
			  m->mod_unload();

			  db[ i ].ptr->mod_unload();
			  db.erase( db.begin() + i, db.begin() + i + 1 );
			  
			  return( 0 );
		  }
	 }

	return( -1 );
}

/// Beendet alle Module.
///
/// @return 0, wenn kein Fehler auftrat. Sonst -1.
int module_master::unload_all( void )
{
	for( int i = 0; i < db.size(); i++ )
	{
		if( db[ i ].ptr )
			db[ i ].ptr->mod_unload();
	}

	return( 0 );
}

/// Reicht eine Nachricht an alle Module weiter. Wie diese darauf reagieren,
/// ist in handle_msg festgelegt.
///
/// @param msg Die zu sendende Nachricht
///
/// @return 0, wenn kein Fehler auftrat. Sonst -1.
int module_master::broadcast( const char* msg )
{
	for( int i = 0; i < db.size(); i++ )
		db[ i ].ptr->handle_msg( msg );

	return( 0 );
}

/// Liefert die Adresse eines bestimmten Moduls. Wenn eine ID von -1 angegeben wird,
/// was standardmig der Fall ist, geht das Programm davon aus, dass das erste
/// Modul, das den selben Namen trgt, gefunden werden soll.
///
/// @param name Der Name des Moduls, beispielsweise "module_net_tcp"
/// @param id Die ID des Moduls. Momentan ungenutzt.
///
/// @return Zeiger auf Modul oder NULL, wenn dieses nicht gefunden wurde.
module* module_master::get_module( const char* name, int id )
{
	for( int i = 0; i < db.size(); i++ )
	{
		if( 	strstr( db[ i ].name, name ) &&
			( db[ i ].id == id || id == -1 ) )
			return( db[ i ].ptr );
	}

	return( NULL );
}
