/*
	~~~~~~~~~~~~~~~~~~
	~ module_map.cpp ~
	~~~~~~~~~~~~~~~~~~

Modul zur Simulation der Roboterbewegungen. Dient zudem als Datenspeicher
fr die Karte. 

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/

// TODO:
// - parse_datalog berprfen

#include <stdio.h>
#include <memory.h>
#include <math.h>
#include "sim_language.h"

#include "module_map.h"

/// Der Konstruktor eines Pfades bentigt nur die zu verbindenden Punkte.
/// Alle anderen Daten werden "automagisch" hinzugefgt.
///
/// HINWEIS: z-Werte werden momentan nicht verwendet.
///
/// @param x1 x-Wert des Startposition.
/// @param y1 y-Wert des Startposition.
/// @param z1 z-Wert des Startposition.
/// @param x2 x-Wert der Zielposition.
/// @param y2 y-Wert der Zielposition.
/// @param z2 z-Wert der Zielposition.
/// @param permanent Gibt an, ob der Pfad permanent sein soll.
robot_path::robot_path( double x1, double y1, double z1, double x2, double y2, double z2, bool permanent )
{
	source_x = x1;
	source_y = y1;
	source_z = z1;

	target_x = x2;
	target_y = y2;
	target_z = z2;
	
	this->permanent = permanent;
	active = true;
	gettimeofday( &creation_time, NULL );
}

/// Gibt den Zeitpunkt der "Erschaffung" der Klasse zurck.
///
/// @return Eine timeval-Struktur, die die Daten von creation_time
/// enthlt.
timeval robot_path::get_creation_time( void )
{
	return( creation_time );
}

/// Setzt die Permanenzeigenschaft eines Pfades. Permanente
/// Pfade werden nicht nach einer gewissen Zeit gelscht.
///
/// @param permanent Gibt an, ob der Pfad transparent sein soll
void robot_path::set_permanent( bool permanent )
{
	this->permanent = permanent;
}

/// Gibt zurck, ob es sich um einen permanenten Pfad
/// handelt oder nicht.
///
/// @return True, wenn der Pfad permanent ist. Sonst false.
bool robot_path::get_permanent( void )
{
	return( permanent );
}

/// Setzt den Status des Pfades. Dieser kann entweder aktiv sein, sodass
/// er noch von der Karte gezeichnet wird, oder inaktiv.
///
/// @param active Gibt an, ob der Pfad aktiv ist oder nicht.
void robot_path::set_status( bool active )
{
	this->active = active;
}

/// Gibt den Status des Pfades zurck.
///
/// @return True, wenn der Pfad noch aktiv ist. Sonst false.
bool robot_path::get_status( void )
{
	return( active );
}

/// Gibt den x-Wert der Startposition zurck.
double robot_path::get_source_x( void )
{
	return( source_x );
}

/// Gibt den y-Wert der Startposition zurck.
double robot_path::get_source_y( void )
{
	return( source_y );
}

/// Gibt den z-Wert der Startposition zurck. Momentan nicht verwendet.
double robot_path::get_source_z( void )
{
	return( source_z );
}

/// Gibt den x-Wert der Zielposition zurck.
double robot_path::get_target_x( void )
{
	return( target_x );
}

/// Gibt den y-Wert der Zielposition zurck.
double robot_path::get_target_y( void )
{
	return( target_y );
}

/// Gibt den z-Wert der Zielposition zurck.
double robot_path::get_target_z( void )
{
	return( source_x );
}


/// Der Konstruktor bereitet eine leere Karte mit Standardeinstellungen vor.
/// ber die Funktion setup() kann die Karte schlielich erzeugt werden.
module_map::module_map( void )
{
	height = width = 0;
	map_data.clear();
	paths.clear();
	robot.x = robot.y = robot.z = 0.0;
	robot.angle = 0.0;
   	robot.sigma_x = robot.sigma_y = robot.sigma_z = 0.0;
	robot.sigma_angle = 0.0;
}

/// Beim Laden des Moduls wird die Karte zunchst noch nicht intialisiert.
/// Dies kann zu einem spteren Zeitpunkt durch die Funktion setup() erledigt
/// werden.
///
/// mod_load() sorgt dafr, dass alle bentigten Variablen aus der Konfigurationsdatei
/// gelesen werden.
///
/// @return 0, wenn das Laden erfolgreich war. Sonst -1.
int module_map::mod_load( void )
{
	mod_conf = dynamic_cast<module_conf*>( module_master::ptr().get_module( "module_conf" ) );
	if( mod_conf == NULL )
	{
		module_master::ptr().broadcast( "*module_map: Kein Konfigurationsmodul gefunden! Fatal!\n" );
		return( -1 );
	}

	path_ttl = atoi( mod_conf->da_datum( "PFAD_LEBENSDAUER" ) );

	BATTERIESPANNUNG = static_cast<unsigned int>( atoi( mod_conf->da_datum( "BATTERIESPANNUNG" ) ) );
	UNGUELTIGER_BEFEHL = static_cast<unsigned int>( atoi( mod_conf->da_datum( "UNGUELTIGER_BEFEHL" ) ) );
	BEFEHL_VERWEIGERT = static_cast<unsigned int>( atoi( mod_conf->da_datum( "BEFEHL_VERWEIGERT" ) ) );
	FATALER_FEHLER = static_cast<unsigned int>( atoi( mod_conf->da_datum( "FATALER_FEHLER" ) ) );
	AUSFUEHRUNGSPROTOKOLL_ANFANG = static_cast<unsigned int>( atoi( mod_conf->da_datum( "AUSFUEHRUNGSPROTOKOLL_ANFANG" ) ) );
	AUSFUEHRUNGSPROTOKOLL_ENDE = static_cast<unsigned int>( atoi( mod_conf->da_datum( "AUSFUEHRUNGSPROTOKOLL_ENDE" ) ) );
	METERZEIT = static_cast<unsigned int>( atoi( mod_conf->da_datum( "METERZEIT" ) ) );
	PERIODENDAUER = static_cast<unsigned int>( atoi( mod_conf->da_datum( "PERIODENDAUER" ) ) );
	GENAUIGKEIT = static_cast<unsigned int>( atoi( mod_conf->da_datum( "GENAUIGKEIT" ) ) );
	SENSOR_1_AN = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_1_AN" ) ) );
	SENSOR_2_AN = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_2_AN" ) ) );
	SENSOR_3_AN = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_3_AN" ) ) );
	SENSOR_1_AUS = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_1_AUS" ) ) );
	SENSOR_2_AUS = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_2_AUS" ) ) );
	SENSOR_3_AUS = static_cast<unsigned int>( atoi( mod_conf->da_datum( "SENSOR_3_AUS" ) ) );

	EPM = atof( mod_conf->da_datum( "EPM" ) );
	
	SENSOR_1_X = atof( mod_conf->da_datum( "SENSOR_1_X" ) );
	SENSOR_1_Y = atof( mod_conf->da_datum( "SENSOR_1_Y" ) );
	SENSOR_1_RTG_X = atof( mod_conf->da_datum( "SENSOR_1_RTG_X" ) );
	SENSOR_1_RTG_Y = atof( mod_conf->da_datum( "SENSOR_1_RTG_Y" ) );
	SENSOR_1_BREITE = atof( mod_conf->da_datum( "SENSOR_1_BREITE" ) );

	SENSOR_2_X = atof( mod_conf->da_datum( "SENSOR_2_X" ) );
	SENSOR_2_Y = atof( mod_conf->da_datum( "SENSOR_2_Y" ) );
	SENSOR_2_RTG_X = atof( mod_conf->da_datum( "SENSOR_2_RTG_X" ) );
	SENSOR_2_RTG_Y = atof( mod_conf->da_datum( "SENSOR_2_RTG_Y" ) );
	SENSOR_2_BREITE = atof( mod_conf->da_datum( "SENSOR_2_BREITE" ) );

	SENSOR_3_X = atof( mod_conf->da_datum( "SENSOR_3_X" ) );
	SENSOR_3_Y = atof( mod_conf->da_datum( "SENSOR_3_Y" ) );
	SENSOR_3_RTG_X = atof( mod_conf->da_datum( "SENSOR_3_RTG_X" ) );
	SENSOR_3_RTG_Y = atof( mod_conf->da_datum( "SENSOR_3_RTG_Y" ) );
	SENSOR_3_BREITE = atof( mod_conf->da_datum( "SENSOR_3_BREITE" ) );

	battery_voltage = 8000.0;
	return( 0 );
}

/// Am Ende des Daseins dieses Moduls muss nur der Speicher freigegeben
/// werden.
///
/// @return Immer 0, da nichts schiefgehen kann.
int module_map::mod_unload( void )
{
	map_data.clear();
	paths.clear();
	return( 0 );	
}

/// Der Status ist hier 0, wenn die Karte bereits intialisiert wurde,
/// und -1, wenn dem nicht so ist.
///
/// @return 0 oder -1.
int module_map::get_status( void )
{
	// Der besseren Lesbarkeit halber ist die Bedingung
	// hier ausgeschrieben.
	if( ( height > 0 )  && ( width > 0 ) )
		return( 0 );
	else
		return( -1 );	
}

/// Das Modul reagiert in dieser Version nicht auf Nachrichten.
///
/// @return 0, da keine Behandlung von Nachrichten existiert.
int module_map::handle_msg( const char* msg )
{
	return( 0 );
}

/// Erstellt eine neue Karte. width und height werden zunchst nur
/// dafr verwendet, zu berprfen, ob hinzugefgte Objekte auf dem
/// Kartenausschnitt dargestellt werden knnen.
///
/// @param width Die Breite der Karte (jeweils ein Kartenelement).
/// @param height Die Hhe der Karte (jeweils ein Kartenelement).
///
/// @return 0, wenn es keinen Fehler gab. Sonst -1. 
int module_map::setup( double width, double height )
{
	map_data.clear();
		
	this->width = width;
	this->height = height;

	robot.x = width / 2;
   	robot.sigma_x = 0.0;
	robot.y = height / 2;
	robot.sigma_y = 0.0;
	robot.z = 0.0;
	robot.sigma_z = 0.0;
	robot.angle = 0;
   	robot.name = (string)(mod_conf->da_datum("ROBOTERNAME"));

	return( 0 );
}

/// Diese berladene Version von setup() liest die Kartendaten aus einer
/// Kartendatei. Die Datei enthlt "width" Spalten und "height" Zeilen, wobei
/// an jeder Stelle der Typ des Terrains steht.
///
/// Gltige Terraintypen sind:
/// - " " (Leerzeichen bzw. Blank) fr TERRAIN_UNKNOWN
/// - "#" (Gatter bzw. Raute) fr TERRAIN_BLOCKED_UNKNOWN
///
///
/// @param width Breite der Karte.
/// @param height Hhe der Karte.
/// @param map_file ASCII-Datei, die eine Beschreibung der Karte enthlt.
///
/// @return 0, wenn es keinen Fehler gab. Sonst -1. 
int module_map::setup( double width, double height, const char* map_file )
{
	setup( width, height );
	
	FILE* data;
	if( ( data = fopen( map_file, "r" ) ) == NULL )
	{
		return( -1 );
	}

	// Jedes Zeichen in einer Zeile wird als x-Wert interpretiert,
	// wobei die Nummer der Zeile den y-Wert angibt.
	
	char* buf = new char[ static_cast<unsigned int>( width ) * 2 + 1];
	unsigned int y = static_cast<unsigned int>( height ) - 1;
	unsigned int type = 0;

	map_info cur_map_element;
	while( !feof( data ) && y >= 0 )
	{
		memset( buf, 0, static_cast<unsigned int>( width ) * 2 + 1 );
		fgets( buf, static_cast<unsigned int>( width ) * 2 + 1, data );

		if( strlen( buf ) < width )
		{
			module_master::ptr().broadcast( "*module_map: Zeilenl�ge nicht gro�genug.\n" );
			break;
		}
		
		// Falls wir auf ein unbekanntes Zeichen stoen sollten,
		// interpretieren wir dieses als TERRAIN_UNKNOWN. Passt ja
		// auch irgendwie.

		for( unsigned int x = 0; x < width; x++ )
		{
			switch( buf[ x ] )
			{
				case ' ':
					type = TERRAIN_UNKNOWN;
					break;
				case '#':
					type = TERRAIN_BLOCKED_UNKNOWN;
					break;
				
				default:
					type = TERRAIN_UNKNOWN;		
			};

			if( type == TERRAIN_UNKNOWN )
				continue;

			cur_map_element.x = x;
			cur_map_element.y = y;
			cur_map_element.z = 0.0;

			cur_map_element.type = type;
			cur_map_element.angle = 0.0;
			
			map_data.push_back( cur_map_element );
		}

		if( y == 0 )
			break;	
		y--;
	}

	delete[] buf;
	fclose( data );

	return( 0 );
}

/// Gibt die bisher erkannten Hindernisse in einem Vektor zurck.
///
/// @return Ein Vektor vom Typ map_info, der alle Hindernisse, auf
/// die der Roboter bereits gestoen ist, enthlt.
vector<map_info> module_map::get_map_data( void )
{
	return( map_data );
}

/// Gibt die Breite der Karte zurck.
///
/// @return Breite der Karte (> 0) oder 0, falls keine Karte
/// intialisiert wurde.
double module_map::get_width( void )
{
	return( width );
}

/// Gibt die Hhe der Karte zurck.
///
/// @return Hhe der Karte (> 0) oder 0, falls keine Karte
/// initialisiert wurde.
double module_map::get_height( void )
{
	return( height );
}

/// Gibt die gespeicherten Pfade des Roboters als STL-Vektor
/// zurck. Diese Daten knnen spter fr andere Auswertungen
/// verwendet werden.
///
/// @return Ein Vektor vom Typ robot_path.
vector <robot_path> module_map::get_paths( void )
{
	return( paths );
}

/// Gibt die aktuell gespeicherten Daten ber den Roboter zurck. Diese Funktion
/// wird zum Beispiel beim Zeichnen des OpenGL-Fensters bentigt.
///
/// @return Ein Objekt vom Typ robot_data.
robot_data module_map::get_robot_data( void )
{
	return( robot );
}

/// Fgt unbekannte Kommandos in eine STL-Map ein, sodass das
/// Modell wei, welcher Code zu welchem Befehl, den der Benutzer
/// geben kann, gehrt.
///
/// @param name Der Name des Kommandos, beispielsweise "Vorwaerts".
/// @param code Der Code des Kommandos, beispielsweise 100.
void module_map::add_command( string name, unsigned char code )
{
	if( command_library.find( code ) == command_library.end() )
		command_library[ code ] = name;
}

/// Gibt das Kommando zurck, das zu einem Code gehrt.
///
/// @param code Der Code, nach dem gesucht werden soll.
/// @return Ein String, der im Fall, dass das Kommando nicht
/// gefunden werden konnte, leer ist. Sonst enthlt er
/// das Kommando, das zu dem angegebenen Code gehrt.
string module_map::get_command( unsigned char code )
{
	string empty;
	if( command_library.find( code ) == command_library.end() )
		return( empty );
	else
		return( command_library[ code ] );	
}

/// Fgt ein Kartenelement zu den bereits gespeicherten Elementen
/// hinzu.
///
/// @param element Das hinzuzufgende Element
void module_map::add_map_element( map_info element )
{
	map_data.push_back( element );
}

/// Fgt zu den bereits gespeicherten Pfaden einen neuen hinzu. Dieser darf
/// _nicht_ ber die Begrenzungen der Karte hinausgehen. Dazu werden die Variablen
/// width und height der Karte berprft.
///
/// @param path Der hinzuzufgende Pfad.
/// @return Die Anzahl der bereits gespeicherten Pfade oder 0, falls ein 
/// Fehler aufgetreten ist.
size_t module_map::add_path( robot_path path )
{
	if(		path.source_x > width 
		|| 	path.source_x < 0.0
		||	path.source_y > height
		||	path.source_y < 0.0
		||	path.target_x > width
		||	path.target_x < 0.0
		||	path.target_y > height
		||	path.target_y < 0.0 )
	{
		module_master::ptr().broadcast( "*module_map: In add_path: Konnte Pfad wegen falscher Koordinaten nicht hinzufgen.\n" );
		return( 0 );
	}

	paths.push_back( path );
	return( paths.size() );
}

/// Geht das Datalog, das der Roboter nach Ausfhrung der Befehle schickt, durch,
/// und versucht, die entsprechenden Kommandos auszufhren. Siehe dazu auch die
/// Erluterungen in der Dokumentation.
///
/// @param saved_messages Die ursprnglich vom Benutzer versandten Nachrichten.
/// @param datalog Das Datalog, das der Roboter nach Ausfhrung aller Befehle
/// angelegt hat. Kann auch unvollstndig sein, wenn beispielsweise eine Kollision
/// aufgetreten ist.
void module_map::parse_datalog( vector<saved_net_msg> saved_messages, string datalog )
{
	// TODO:
	// - Was passiert, wenn man einen leeren String ("") an einen nicht leeren
	// String anhngt und dann c_str() aufruft?
	//
	// - Der 'else'-Fall in der while-Schleife drfte eigentlich nicht auftreten,
	// wenn der Roboter alle Kommandos empfangen hat.

	string temp;
	if( atoi( datalog.substr( 0, datalog.find_first_of( "\n" ) ).c_str() ) != BATTERIESPANNUNG )
	{
		module_master::ptr().broadcast( "*module_map: In parse_datalog: Fehlerhaftes Datalog empfangen (Batteriespannung fehlt):\n" );
		module_master::ptr().broadcast( ( "*** Anfang des Datalogs **\n" + datalog + "\n" ).c_str() );
		module_master::ptr().broadcast( "*** Ende des Datalogs **\n" );
		

		return;
	}

	// Batteriespannung lesen und setzen
	
	datalog.erase( 0, datalog.find_first_of( "\n" )+ 1 );
	temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );

	if( temp != "" )
		set_battery_voltage( atoi( temp.c_str() ) );
	else
	{
		module_master::ptr().broadcast( "*module_map: In parse_datalog: Fehlerhaftes Datalog empfangen (Batteriespannung fehlt):\n" );
		module_master::ptr().broadcast( ( "*** Anfang des Datalogs **\n" + datalog + "\n" ).c_str() );
		module_master::ptr().broadcast( "*** Ende des Datalogs **\n" );
		

		return;
	}	

	// Auf gltiges Datalog testen.

	datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
	temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );

	if( temp == "" )
	{
		module_master::ptr().broadcast( "*module_map: In parse_datalog: Fehlerhaftes Datalog empfangen (Batteriespannung fehlt):\n" );
		module_master::ptr().broadcast( ( "*** Anfang des Datalogs **\n" + datalog + "\n" ).c_str() );
		module_master::ptr().broadcast( "*** Ende des Datalogs **\n" );
		
		return;
	}

	else if( atoi( temp.c_str() ) != AUSFUEHRUNGSPROTOKOLL_ANFANG )
	{
	
		module_master::ptr().broadcast( "*module_map: In parse_datalog: Fehlerhaftes Datalog empfangen (kein Ausfhrungsprotokoll):\n" );
		module_master::ptr().broadcast( ( "*** Anfang des Datalogs **\n" + datalog + "\n" ).c_str() );
		module_master::ptr().broadcast( "*** Ende des Datalogs **\n" );
		
		return;
	}
	
	datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
	
	// Ab hier liegt nun das Datalog in einer Form vor, die entweder Sensordaten
	// oder ausgefhrte Kommandos enthlt.

	unsigned int cmd_num = 0;
	unsigned int temp_cmd = 0;
	unsigned int time = 0; // Die Zeit bei Kollisionen (in Dezisekunden)

	// Bettigte Sensoren rausfinden

	bool sensor_1 = ( datalog.find( mod_conf->da_datum( "SENSOR_1_AN" ) ) != string::npos );
	bool sensor_2 = ( datalog.find( mod_conf->da_datum( "SENSOR_2_AN" ) ) != string::npos );
	bool sensor_3 = ( datalog.find( mod_conf->da_datum( "SENSOR_3_AN" ) ) != string::npos );

	while( datalog.find_first_of( "\n" ) != string::npos && cmd_num < saved_messages.size() )
	{
		temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );
		temp_cmd = static_cast<unsigned int>( atoi( temp.c_str() ) );
	
		// Das gelesene Kommando stimmt mit dem gespeicherten Kommando berein. Nun wird noch
		// die nchste Zeile berprft, ob vielleicht ein Sensorereignis vorliegt.	
		if( temp_cmd == saved_messages[ cmd_num ].header.get_cmd() )
		{
			datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
			temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );

			// Kein Sensor wurde ausgelst, das Kommando kann also ausgefhrt
			// werden.
			if( 	static_cast<unsigned int>( atoi( temp.c_str() ) ) != SENSOR_1_AN &&
				static_cast<unsigned int>( atoi( temp.c_str() ) ) != SENSOR_2_AN &&
 				static_cast<unsigned int>( atoi( temp.c_str() ) ) != SENSOR_3_AN )
			{
				execute_command( temp_cmd, saved_messages[ cmd_num ] );
				cmd_num++;

				continue;
			}

			// Sensorereignis; die Ausfhrungszeit muss neu berechnet werden.
			else
			{
				 datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
				 time = atoi( datalog.substr( 0, datalog.find_first_of( "\n" ) ).c_str() );
				 break;
			}
			
		}
		else if( temp_cmd == AUSFUEHRUNGSPROTOKOLL_ENDE )
			break;
		else if( temp_cmd == BEFEHL_VERWEIGERT )
		{
			// Verweigerten Befehl lesen 
			
			datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
			temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );

			module_master::ptr().broadcast(  static_cast<string>( "#module_map: Der Roboter hat einen Befehl verweigert: " + temp + "\n" ).c_str() );
		}
		else if( temp_cmd == UNGUELTIGER_BEFEHL )
		{
			// Ungltigen Befehl lesen
			datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
			temp = datalog.substr( 0, datalog.find_first_of( "\n" ) );

			module_master::ptr().broadcast(  static_cast<string>( "#module_map: Der Roboter hat einen ungltigen Befehl gemeldet: " + temp + "\n" ).c_str() );
		}
		else
		{
			fprintf( 	stderr, 
					"*** MODULE_MAP ***\n",
					"%s\n"
					"******************\n", datalog.c_str() );
						
			module_master::ptr().broadcast( "#module_map: In parse_datalog: Der Roboter konnte nicht alle Kommandos ababeiten.\n" );
			cmd_num++;
		}

		datalog.erase( 0, datalog.find_first_of( "\n" ) + 1 );
	}

		
	// Nur ausfhren, wenn keine Sensoren aktiviert wurden
	if( 	!sensor_1 &&
		!sensor_2 &&
		!sensor_3 )
		return;

	fprintf( stderr, "** SENSOREREIGNIS **" );

	// saved_messages[ cmd_num ] enthlt das Kommando, das ausgefhrt
	// wurde, whrend der Sensorkontakt stattfand.
		
	vector<unsigned char> params_vector;	
	vector<conf_command> parsed_commands;
	string user_command = get_command( saved_messages[ cmd_num ].header.get_cmd() );

	string user_command_elementary = mod_conf->da_datum( (static_cast<string>( "CMD_" ) + user_command).c_str() );
	parsed_commands = term_parse_cmd( user_command_elementary );

	// Parametervektor aufstellen
	for( size_t i = 0; i < saved_messages[ cmd_num ].params.size(); i++ )
		params_vector.push_back( static_cast<unsigned char>( saved_messages[ cmd_num ].params[ i ] ) );

	// Drehfaktoren anhand der neuen Spannung berechnen
	
	vector<double> batt_voltage;
	batt_voltage.push_back( battery_voltage );
	
	double gehfaktor = ((term_terminate((string)(mod_conf->da_datum("METERZEIT")))[0].precisely_evaluate(batt_voltage))/EPM);
	double drehfaktor = ( ((term_terminate((string)(mod_conf->da_datum("PERIODENDAUER")))[0].precisely_evaluate(batt_voltage))/360.0) );

	// Zeiten schtzen
	
	double cmd_execution_time = 0.0;
	double eh_faktor = 0.0; // Ist entweder der Drehfaktor oder der Gehfaktor

	for( size_t i = 0; i < parsed_commands.size(); i++ )
	{
		if( 	parsed_commands[ i ].name == "VORWAERTS" ||
			parsed_commands[ i ].name == "RUECKWAERTS" )
		{
			cmd_execution_time += gehfaktor * parsed_commands[ i ].formula.evaluate( params_vector );
			eh_faktor = gehfaktor;
		}
		else if( 	parsed_commands[ i ].name == "LINKS" ||
			 	parsed_commands[ i ].name == "RECHTS" )
		{
			cmd_execution_time += drehfaktor * parsed_commands[ i ].formula.evaluate( params_vector );
			eh_faktor = drehfaktor;
		}
		
		// Ausfhrung des Kommandos, wenn es vor dem Sensorkontakt beendet war
		if( ( cmd_execution_time / 10.0 ) < static_cast<double>( time ) )
			execute_command( parsed_commands[ i ].name, parsed_commands[ i ].formula.evaluate( params_vector ) );
		else
		{
			cmd_execution_time -= eh_faktor * parsed_commands[ i ].formula.evaluate( params_vector );
			execute_command( parsed_commands[ i ].name, ( - cmd_execution_time  + time * 10 ) / ( eh_faktor ) );

			// Hindernisse zeichnen
			
			map_info obstacle;
			double sensor_angle = 0.0;
			double _sin = sin( robot.angle * PI / 180.0 );
			double _cos = cos( robot.angle * PI / 180.0 );

			if( sensor_1 )
			{
				obstacle.x = _cos * SENSOR_1_X - _sin * SENSOR_1_Y + robot.x;
				obstacle.y = _sin * SENSOR_1_X + _cos * SENSOR_1_Y + robot.y;
				obstacle.z = 0.0;
				obstacle.width = SENSOR_1_BREITE;
				
				sensor_angle = acos( SENSOR_1_RTG_Y / sqrt( SENSOR_1_X * SENSOR_1_X + SENSOR_1_Y * SENSOR_1_Y ) ) * 180.0 /PI;
				obstacle.angle = sensor_angle + robot.angle;

				add_map_element( obstacle );
			}

			if( sensor_2 )
			{
				obstacle.x = _cos * SENSOR_2_X - _sin * SENSOR_2_Y + robot.x;
				obstacle.y = _sin * SENSOR_2_X + _cos * SENSOR_2_Y + robot.y;
				obstacle.z = 0.0;
				obstacle.width = SENSOR_2_BREITE;
				
				sensor_angle = acos( SENSOR_2_RTG_Y / sqrt( SENSOR_2_X * SENSOR_2_X + SENSOR_2_Y * SENSOR_2_Y ) ) * 180.0 /PI;
				obstacle.angle = sensor_angle + robot.angle;
				
				add_map_element( obstacle );
			}

			if( sensor_3 )
			{
				obstacle.x = _cos * SENSOR_3_X - _sin * SENSOR_3_Y + robot.x;
				obstacle.y = _sin * SENSOR_3_X + _cos * SENSOR_3_Y + robot.y;
				obstacle.z = 0.0;
				obstacle.width = SENSOR_3_BREITE;
	
				sensor_angle = acos( SENSOR_3_RTG_Y / sqrt( SENSOR_3_X * SENSOR_3_X + SENSOR_3_Y * SENSOR_3_Y ) ) * 180.0 /PI;
				obstacle.angle = sensor_angle + robot.angle;
				
				add_map_element( obstacle );
			}
		}
	}
}

/// Schtzt die Zeit, die zur Ausfhrung aller angegebenen Befehle bentigt wird,
/// ab, und schickt sie an den Server.
///
/// @param messages Die vom Benutzer eingegebenen Befehle. Diese sollen an den Roboter
/// gesandt werden.
///
/// @return Eine Zeit in Sekunden, die ber MSG_EXECUTION_TIME an den Server
/// bertragen wird.
double module_map::estimate_execution_time( vector<saved_net_msg> messages )
{
	vector<unsigned char> params_vector;
	vector<double> batt_voltage;
	string user_command;
	string user_command_elementary;	// Der Befehl des Benutzers in Elementarbefehle zerlegt
	vector<conf_command> parsed_commands;
	double parsed_command_param = 0.0;
	double time = 0.0;

	// Drehfaktoren anhand der neuen Spannung berechnen

	batt_voltage.push_back( battery_voltage );
	
	double gehfaktor = ((term_terminate((string)(mod_conf->da_datum("METERZEIT")))[0].precisely_evaluate(batt_voltage))/EPM);
	double drehfaktor = ( ((term_terminate((string)(mod_conf->da_datum("PERIODENDAUER")))[0].precisely_evaluate(batt_voltage))/360.0) );


	for( size_t i = 0; i < messages.size(); i++ )
	{
		params_vector.clear();
		parsed_commands.clear();
		
		user_command = get_command( messages[ i ].header.get_cmd() );
		user_command_elementary = mod_conf->da_datum( (static_cast<string>( "CMD_" ) + user_command).c_str() );
	
		if( user_command_elementary == "" )
			return( 0.0 );
		
		parsed_commands = term_parse_cmd( user_command_elementary );
		
		// Parametervektor aufstellen
		for( size_t j = 0; j < messages[ i ].params.size(); j++ )
			params_vector.push_back( static_cast<unsigned char>( messages[ i ].params[ j ] ) );

		// Zeiten schtzen
		for( size_t j = 0; j < parsed_commands.size(); j++ )
		{
			parsed_command_param = parsed_commands[ i ].formula.evaluate( params_vector );
			
			// Faktor draufpacken
			if( 	parsed_commands[ j ].name == "VORWAERTS" ||
				parsed_commands[ j ].name == "RUECKWAERTS" )
				parsed_command_param *= static_cast<double>( gehfaktor );
			else if( 	parsed_commands[ j ].name == "LINKS" ||
				 	parsed_commands[ j ].name == "RECHTS" )
				parsed_command_param *= static_cast<double>( drehfaktor ); 
				
			time += parsed_command_param;
		}
	}

	return( time / 100.0  );
}

/// Fhrt ein Kommando aus, das der Benutzer dem Roboter geschickt hat. Die
/// Karte zerlegt das Kommando dann in elementare Bewegungen wie zum Beispiel
/// "Vorwrts". Mehr Informationen sind im Benutzerhandbuch zu finden.
/// 
///
/// @param code Der Bytecode, der zum angegebenen Befehl gehrt.
/// @param message Der vom Benutzer eingegebene Befehl mit Parametern. Der
/// tatschliche Parameter fr den bergebenen Elementarbefehl wird erst
/// noch berechnet.
void module_map::execute_command( unsigned char code, const saved_net_msg& message )
{
	vector<unsigned char> params_vector;
	string user_command;
	string user_command_elementary;	// Der Befehl des Benutzers in Elementarbefehle zerlegt
	vector<conf_command> parsed_commands;
	int parsed_command_param = 0;

	params_vector.clear();
		
	user_command = get_command( message.header.get_cmd() );
	user_command_elementary = mod_conf->da_datum( (static_cast<string>( "CMD_" ) + user_command).c_str() );
		
	parsed_commands = term_parse_cmd( user_command_elementary );
		
	// Parametervektor aufstellen
	for( size_t i = 0; i < message.params.size(); i++ )
		params_vector.push_back( static_cast<unsigned char>( message.params[ i ] ) );

	// Kommandos ausfhren, so mglich
	for( size_t i = 0; i < parsed_commands.size(); i++ )
	{
		parsed_command_param = parsed_commands[ i ].formula.evaluate( params_vector );
		execute_command( parsed_commands[ i ].name, parsed_command_param );
	}
}

/// Fhrt ein Elementarkommando wie beispielsweise eine Bewegung nach vorne
/// aus und stellt das Ergebnis auf der Karte dar. Im Unterschied zur anderen
/// execute_command-Funktion ist der Parameter des Kommandos bekannt und muss
/// nicht erst berechnet werden.
///
/// @param command Das auszufhrende Kommando. Momentan werden die Befehle
/// Vorwrts, Rckwrts, Linksdrehung und Rechtsdrehung untersttzt..
/// @param param Der Parameter, mit dem das Kommando aufgerufen wird. Dieser
/// Wert gibt beispielsweise an, wie weit sich der Roboter bewegen soll. Eventuell
/// muss der Parameter noch berechnet werden. Hierfr existiert eine andere Version
/// von execute_command.
///
/// Der Parameter sollte double sein, damit partiell ausgefhrte Kommandos
/// auf der Karte richtig dargestellt werden knnen.
void module_map::execute_command( string command, double param )
{
	transform( command.begin(), command.end(), command.begin(), (int(*)(int)) toupper);
	
	robot_data new_robot_data;
	new_robot_data = robot;
   
   	vector<double> temp;
	temp.clear();
	temp.push_back((double)param);
	temp.push_back( get_battery_voltage() );   
	
	if( command  == "VORWAERTS" )
	{
		new_robot_data.x += - param * sin( new_robot_data.angle * PI / 180.0 );
		new_robot_data.y += param * cos( new_robot_data.angle * PI / 180.0 );
      //Gaußsche Fehlerfortpflanzung: (Hoffe ich.)
      new_robot_data.sigma_x = sqrt(pow(robot.sigma_x, 2.0) 
                                    + pow(cos(robot.angle * PI / 180.0 )
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_VORWAERTS")))[0].precisely_evaluate(temp)
                                          , 2.0)
                                    + pow((double)param * sin(robot.angle * PI / 180.0 )
                                          * robot.sigma_angle * PI / 180.0 , 2.0));
      new_robot_data.sigma_y = sqrt(pow(robot.sigma_y, 2.0) 
                                    + pow(sin(robot.angle * PI / 180.0 )
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_VORWAERTS")))[0].precisely_evaluate(temp)
                                          , 2.0)
                                    + pow((double)param * cos(robot.angle * PI / 180.0 )
                                          * robot.sigma_angle * PI / 180.0 , 2.0));
      new_robot_data.sigma_angle = sqrt(pow(robot.sigma_angle * PI / 180.0 , 2.0)
                                        * pow((term_terminate((string)(mod_conf->da_datum("SIGMA_WINKEL_VORWAERTS")))[0].precisely_evaluate(temp)
                                               * PI / 180.0)                                        
                                              , 2.0))
                                    / PI * 180.0 ; //in Grad zurueckverwandeln.
	}
	
	else if ( command == "RUECKWAERTS" )
	{
		new_robot_data.x += param * sin( new_robot_data.angle * PI / 180.0 );
		new_robot_data.y -= param * cos( new_robot_data.angle * PI / 180.0 );
      //Gaußsche Fehlerfortpflanzung:
      new_robot_data.sigma_x = sqrt(pow(robot.sigma_x, 2.0) 
                                    + pow(sin(robot.angle * PI / 180.0 )
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_RUECKWAERTS")))[0].precisely_evaluate(temp)
                                          , 2.0)
                                    + pow((double)param * cos(robot.angle * PI / 180.0 )
                                          * robot.sigma_angle * PI / 180.0 , 2.0));
      new_robot_data.sigma_y = sqrt(pow(robot.sigma_y, 2.0) 
                                    + pow(cos(robot.angle * PI / 180.0 )
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_RUECKWAERTS")))[0].precisely_evaluate(temp)
                                          , 2.0)
                                    + pow((double)param * sin(robot.angle * PI / 180.0 )
                                          * robot.sigma_angle * PI / 180.0 , 2.0));
      new_robot_data.sigma_angle = sqrt(pow(robot.sigma_angle * PI / 180.0 , 2.0)
                                        + pow((term_terminate((string)(mod_conf->da_datum("SIGMA_WINKEL_RUECKWAERTS")))[0].precisely_evaluate(temp)
                                               * PI / 180.0)
                                              , 2.0))
                                   / PI * 180.0 ; //in Grad zurueckverwandeln.;
	}
	
	else if( command == "LINKS" )
   	{
		new_robot_data.angle += param;
      //Gaußsche Fehlerfortpflanzung:
      new_robot_data.sigma_x = sqrt(pow(robot.sigma_x, 2.0)
                                    + pow(sin(robot.angle * PI / 180.0)
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_LINKS")))[0].precisely_evaluate(temp)
                                          , 2.0));
      new_robot_data.sigma_y = sqrt(pow(robot.sigma_y, 2.0)
                                    + pow(cos(robot.angle * PI / 180.0)
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_LINKS")))[0].precisely_evaluate(temp)
                                          , 2.0));
      new_robot_data.sigma_angle = sqrt(pow(robot.sigma_angle * PI / 180.0, 2.0)
                                        + pow((term_terminate((string)(mod_conf->da_datum("SIGMA_WINKEL_LINKS")))[0].precisely_evaluate(temp)
                                               * PI / 180.0)
                                              , 2.0))
                                   / PI * 180.0 ; //in Grad zurueckverwandeln.;
   	}
	
	else if( command == "RECHTS" )
   	{
      		new_robot_data.angle -= param;
		//Gaußsche Fehlerfortpflanzung:
		new_robot_data.sigma_x = sqrt(pow(robot.sigma_x, 2.0)
                	                 + pow(sin(robot.angle * PI / 180.0)
                       			 * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_RECHTS")))[0].precisely_evaluate(temp)
                                          , 2.0));
      new_robot_data.sigma_y = sqrt(pow(robot.sigma_y, 2.0)
                                    + pow(cos(robot.angle * PI / 180.0)
                                          * term_terminate((string)(mod_conf->da_datum("SIGMA_STRECKE_RECHTS")))[0].precisely_evaluate(temp)
                                          , 2.0));
      new_robot_data.sigma_angle = sqrt(pow(robot.sigma_angle * PI / 180.0, 2.0)
                                        + pow((term_terminate((string)(mod_conf->da_datum("SIGMA_WINKEL_RECHTS")))[0].precisely_evaluate(temp))
                                               * PI / 180.0
                                              , 2.0))
                                   / PI * 180.0 ; //in Grad zurueckverwandeln.;
   	}
	
	// Unbekanntes Kommando
	else
		module_master::ptr().broadcast( "*module_map: In execute_command: Das Modell kann einen unbekannten Elementarbefehl nicht verarbeiten.\n" );

   	new_robot_data.angle = fmod(new_robot_data.angle, (double)360);
	if( new_robot_data.angle < 0 )
		new_robot_data.angle += 360;
	
	robot_path path( 	robot.x, robot.y, robot.z,
				new_robot_data.x, new_robot_data.y, new_robot_data.z );
	if( add_path( path ) )
	{
		robot = new_robot_data;

		// Grnen Pfad hinzufgen
		path.set_permanent( true );
		add_path( path );
	}
}

/// Gibt die aktuell gespeicherte Batteriespannung zurck.
///
/// @param compile_voltage Ist standardmig false. Wenn der Programmierer
/// beim Aufruf true bergibt, so ist das Ergebnis die Batteriespannung, die
/// zum Kompilierzeitpunkt vorlag.
///
/// @return Ein Wert > 0.0 in mV.
double module_map::get_battery_voltage( bool compile_voltage )
{
	if( !compile_voltage )
		return( battery_voltage );
	else
		return( compile_battery_voltage );	
}

/// Setzt die aktuelle Batteriespannung.
///
/// @param voltage Die neue Spannung in mV > 0.0
/// @param compile_voltage Ist standardmig false. Wenn der Programmierer beim
/// Aufruf true bergibt, so wird die Batteriespannung gesetzt, die zum Kompilierzeitpunkt
/// vorlag.
void module_map::set_battery_voltage( double voltage, bool compile_voltage )
{
	if( voltage > 0.0 )
	{
		if( !compile_voltage )
			battery_voltage = voltage;
		else
			compile_battery_voltage = voltage;	
	}
}

/// Unterzieht die aktuelle Kartendarstellung einem Update. Dies bedeutet,
/// dass zunchst alle Pfade auf Gltigkeit berprft werden mssen. Abgelaufene
/// Pfade werden markiert, sodass sie von der Karte nicht mehr gezeichnet werden.
///
/// Hinweis: Als permanent markierte Pfade werden von diesem Update natrlich 
/// ausgeschlossen. Nach Ablauf der vierfachen Lebenszeit werden inaktive Pfade
/// endgltig gelscht, um den Speicherverbrauch zu minimieren.
void module_map::update( void )
{
	// Die Zeit einmal zu Beginn der Schleife zu messen,
	// ist hinreichend genau, da difftime sowieso die vergangene
	// Zeit in Sekunden angibt.
	timeval now;
	gettimeofday( &now, NULL );

	for( size_t i = 0; i < paths.size(); i++ )
	{
		if( difftime( 	now.tv_sec,  
				paths[ i ].creation_time.tv_sec ) >= path_ttl && !paths[ i ].permanent )
		{
			paths[ i ].active = false;
			
			// Pfad engltig lschen, wenn er lang genug inaktiv war
			if( difftime( 	now.tv_sec,
					paths[ i ].creation_time.tv_sec ) >= 4 * path_ttl )
				paths.erase( paths.begin() + i, paths.begin() + i + 1 );
		}
	}
}

/// Schreibt die gespeicherten Pfade in ein allgemein lesbares Datenformat,
/// das die einzelnen Felder durch einen Tabulator trennt.
///
/// @param filename Die Datei, in der alles gespeichert werden soll.
void module_map::dump_paths( const char* filename )
{
	FILE* out = fopen( filename, "w" );
	for( size_t i = 0; i < paths.size(); i++ )
	{
		fprintf( out, 	"%f\t%f\n%f\t%f\n\n",
				paths[ i ].source_x,
				paths[ i ].source_y,
				paths[ i ].target_x,
				paths[ i ].target_y );
	}

	fclose( out );
}
