/*
	~~~~~~~~~~~~~~~~~~
	~ module_map.cpp ~
	~~~~~~~~~~~~~~~~~~

Modul zur Simulation der Roboterbewegungen. Dient zudem als Datenspeicher
fr die Karte. 

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/

#ifndef MODULE_MAP
#define MODULE_MAP

#include <vector>
#include <string>
#include <map>

#include <time.h>
#include <sys/time.h>

#include "module.h"
#include "module_conf.h"
#include "protocol.h"

/// The primary purpose of the DATA statement is to give names to constants;
/// instead of referring to pi as 3.141592653589793 at every appearance, the
/// variable PI can be given that value with a DATA statement and used instead of
/// the longer form of the constant. This also simplifies modifying the program,
/// should the value of pi change. 
/// -- FORTRAN manual for Xerox Computers
const double PI= 3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679;

/// Wenn sich die Batteriespannung des Roboters um mehr als diesen Wert von der Spannung unterscheidet,
/// die beim Kompilierzeitpunkt vorlag, so schlgt das GUI Alarm.
const double DIFF_VOLTAGE_WARN = 0.4;

using namespace std;

/// Abkrzungen fr die einzelnen Terraintypen der Karte.
enum terrain_types
{
	TERRAIN_UNKNOWN = 0,
	TERRAIN_BLOCKED_UNKNOWN,///< Dieser Fleck auf der Karte wird von 
				///< einem unbekannten Hindernis blockiert. 
	TERRAIN_VISITED		///< Dieser Teil der Karte wurde schon abgegrast und fr begehbar befunden.
};

/// Enthlt Daten ber die Position und die Ausrichtung des Roboters. Es ist denkbar,
/// dass sptere Versionen des Programmes mehrere Roboter steuern. Daher ist hier noch
/// zustzlich ein Bezeichner angegeben.
struct robot_data
{
	string name;				///< Ein Name oder eine Bezeichnung fr den Roboter
	double x, y, z;  			///< Position des Roboters in cartesischen Koordinaten
	double angle; 				///< Winkel (in Grad), in dem der Roboter zu einem gedachten Koordinatensystem steht,
						///< das als Achsen die (x,y)-Achsen der Karte verwendet.

   	double sigma_angle;   			///< Fehler des Roboterwinkels.
	double sigma_x, sigma_y, sigma_z;	///< Fehler der Roboterposition
};

/// Verbindungslinien werden bei Positionsnderungen des Roboters erzeugt und
/// gespeichert. Sie zeigen fr eine gewisse Zeitspanne, die in der
/// Konfigurationsdatei als PFAD_LEBENSDAUER zu setzen ist, den Weg des Roboters an.
///
/// Um die Kommunikation untereinander zu vereinfachen, wurde die Klasse
/// module_map als "befreundet" eingetragen.
class robot_path
{
	public:
		robot_path( double x1, double y1, double z1, double x2, double y2, double z2, bool permanent = false );

		timeval get_creation_time( void );

		void set_status( bool active );
		bool get_status( void );

		void set_permanent( bool permanent );
		bool get_permanent( void );

		double get_source_x( void );
		double get_source_y( void );
		double get_source_z( void );

		double get_target_x( void );
		double get_target_y( void );
		double get_target_z( void );

	private:	
		timeval creation_time;
		double source_x, source_y, source_z;
		double target_x, target_y, target_z;

		bool active;
		bool permanent;
	
	friend class module_map;	
};

/// Speichert Daten ber ein Kartenelement. Dieser Ansatz ist flexibler
/// als das einfache Speichern von Terraintypen in einem int-Array.
struct map_info
{
	double x,y,z;				///< Position des Kartenelements
   	double sigma_x, sigma_y, sigma_z;	///< Fehler der Kartenpunktposition

	unsigned int type; 	///< Bezeichnet den aktuellen Terraintyp, z.B. Hindernis
	double width;		///< Hindernisbreite
	double angle;		///< Der Winkel, den das Hindernis einnimmt
	double sigma_angle;  	///< Fehler des Kartenpunktwinkels
	timeval discovery_time;	///< Zeit, zu der das Hindernis entdeckt wurde, momentan nicht verwendet.
};

/// Die Karte besteht aus Kartenelementen. Diese wiederum enthalten Informationen
/// darber, was sie darstellen (beispielsweise "Hindernis" ). Hier wird
/// noch _kein_ Rendering der Karte vorgenommen, da die Daten unabhngig von den
/// verwendeten Grafikbibliotheken gespeichert werden.
class module_map : public module
{
	public:

		module_map( void );

		// Geerbte Methoden
		
		virtual int mod_load( void );
		virtual int mod_unload( void );
		virtual int handle_msg( const char* msg );
	
		/// Identifiziert den Typ des Moduls.
		///
		/// @return "module_map"
		virtual char* get_name( void )
		{
			return( "module_map" );
		}

		virtual int get_status( void );

		// Eigene Methoden

		int setup( double width, double height );
		int setup( double width, double height, const char* map_file );

		void execute_command( string command, double param );
		void execute_command( unsigned char code, const saved_net_msg& message );

		void update( void );
		void dump_paths( const char* filename );
		void parse_datalog( vector<saved_net_msg> saved_messages, string datalog );

		double estimate_execution_time( vector<saved_net_msg> messages );

		// Verschiedene get/set-Funktionen

		double get_width( void );
		double get_height( void );

		void set_battery_voltage( double voltage, bool compile_voltage = false );
		double get_battery_voltage( bool compile_voltage = false );

		robot_data get_robot_data( void );
		
		vector<map_info> get_map_data( void );
		void add_map_element( map_info element );

		vector <robot_path> get_paths( void );
		
		size_t add_path( robot_path path );

		void add_command( string name, unsigned char code );
		string get_command( unsigned char code );

	private:
		double EPM;

		unsigned int BATTERIESPANNUNG;
		unsigned int UNGUELTIGER_BEFEHL;
		unsigned int BEFEHL_VERWEIGERT;
		unsigned int FATALER_FEHLER;
		unsigned int AUSFUEHRUNGSPROTOKOLL_ANFANG;
		unsigned int AUSFUEHRUNGSPROTOKOLL_ENDE;
		unsigned int METERZEIT;
		unsigned int PERIODENDAUER;
		unsigned int GENAUIGKEIT;
		unsigned int SENSOR_1_AN;
		unsigned int SENSOR_2_AN;
		unsigned int SENSOR_3_AN;
		unsigned int SENSOR_1_AUS;
		unsigned int SENSOR_2_AUS;
		unsigned int SENSOR_3_AUS;

		double SENSOR_1_X;
		double SENSOR_1_Y;
		double SENSOR_1_RTG_X;
		double SENSOR_1_RTG_Y;
		double SENSOR_1_BREITE;
		
		double SENSOR_2_X;
		double SENSOR_2_Y;
		double SENSOR_2_RTG_X;
		double SENSOR_2_RTG_Y;
		double SENSOR_2_BREITE;

		double SENSOR_3_X;
		double SENSOR_3_Y;
		double SENSOR_3_RTG_X;
		double SENSOR_3_RTG_Y;
		double SENSOR_3_BREITE;

		module_conf* mod_conf;	///< Zeiger auf das Konfigurationsdateienmodul. Somit knnen Variablen
					///< fr die Karte ausgelesen werden.

		double width; 	///< Aktuelle Breite des Kartenrasters (jeweils ein Kartenelement)
		double height;	///< Aktuelle Hhe des Kartenrasters (jeweils ein Kartenelement)

		double compile_battery_voltage;	///< Die Batteriespannung, die zum Kompilierzeitpunkt vorlag
		double battery_voltage;		///< Aktuelle Batteriespannung
		double path_ttl;		///< Die TTL (Time To Live) eines Pfades in Sekunden. Innerhalb dieser Zeitspanne
						///< wird der Pfad sanft ausgeblendet, sodass er nach deren Ablauf nicht mehr
						///< zu sehen ist.
		vector<map_info> map_data; 	///< Vektor aus Kartenelementen

			
		robot_data robot;
	
		map<unsigned char, string> command_library;	
		vector <robot_path> paths;

		double current_voltage;
};

#endif
