/*
	~~~~~~~~~~~~~~~~~~~~~~
	~ module_net_tcp.cpp ~
	~~~~~~~~~~~~~~~~~~~~~~

Modul zur Netzwerkkommunikation ber TCP/IP. Das Modul kann sowohl einen 
Servermodus wie auch einen Clientmodus bieten.

Dieser Quelltext ist ursprnglich Teil des Programmes Eukalyptos, welches zur
Netzwerkfernsteuerung von Lego-RCX-2-Robotern dient und von Lutz Bch, Bastian
Rieck und Carlos Franke im Rahmen eines Praktikums am Interdisziplinren 
Zentrum fr Wissenschaftliches Rechnen (IWR) der Universitt Heidelberg 
geschaffen wurde.

*/


#include "module_net_tcp.h"
#include "protocol.h"

#include <stdio.h>
#include <memory.h>

unsigned int module_net_tcp::num_instances = 0;

/// Der Konstruktor sorgt nur fr gltige Anfangswerte.
module_net_tcp::module_net_tcp( void )
{
	s = 0;

	is_connected = false;
	is_server = false;
}

/// Initialiserung des Moduls. Momentan nicht ntig.
///
/// @return Immer 0.
int module_net_tcp::mod_load( void )
{
	return( 0 );
}

/// Beenden des Moduls. Schliet u.A. offene Sockets
/// und die setzt die Verbindung zurck.
///
/// @return Immer 0.
int module_net_tcp::mod_unload( void )
{
	if( is_connected )
		disconnect();

	return( 0 );	
}

/// Behandelt Nachrichten. Wenn das Modul ein Server ist und gerade
/// eine Clientverbindung vorliegt, wird eine Netzwerknachricht vom Typ MSG_LOG
/// erstellt, die die empfangene Modulnachricht enthlt und an den
/// Client weiterleitet. Dieser kann sie dann anzeigen.
///
/// @param msg Die empfangene Nachricht
///
/// @return Immer 0.
int module_net_tcp::handle_msg( const char* msg)
{
	if( is_server && is_connected )
	{
		string message = msg;
		net_msg_hdr log_msg( MSG_TYPE_SYSTEM, MSG_LOG, message.size() );
		send_msg( reinterpret_cast<const char*>( &log_msg ), sizeof( net_msg_hdr ), MSG_NOSIGNAL );
		send_msg( message.c_str(), message.size(), MSG_NOSIGNAL );
	}
	return( 0 );
}

/// Erstellt einen Server, der Verbindungen entgegennimmt auf einem
/// bestimmten Port entgegennimmt. Eventuell mssen fr niedrige
/// Ports (< 1024) root-Rechte vorhanden sein.
///
/// @param port Der Port, an den Verbindungen gerichtet werden.
///
/// @return 0, wenn der Server erstellt werden konnte. -1 sonst.
int module_net_tcp::connect( unsigned int port )
{
	int yes = 1;

	// Die Adresse kann weiter verwendet werden, wenn der Server
	// abstrzen sollte.	
	
	s = socket( PF_INET, SOCK_STREAM, 0 );
  	if ( setsockopt( s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof( int ) ) == -1)
	{
		module_master::ptr().broadcast( "*module_net_tcp: Aufruf von Setsockopt ist fehlgeschlagen.\n" );
		return( -1 );
	}

	/// Momentan werden Verbindungen von jeder Adresse
	/// erlaubt. Dies knnte man auch ndern, z.B. auf ein
	/// Subnetz.
	
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_port = htons( port );
	sock_addr.sin_addr.s_addr = INADDR_ANY;
	memset( &(sock_addr.sin_zero), '\0', 8 );

	if( bind( s, (sockaddr*) &sock_addr, sizeof( sockaddr )	)  == - 1 )
	{
		module_master::ptr().broadcast( "*module_net_tcp: Konnte Server nicht an Port binden.\n" ); 
		return( -1 );
	}

	listen( s, MODULE_NET_TCP_MAX_CLIENTS );

	socklen_t client_size = sizeof( struct sockaddr_in );
	if( ( c = accept( s, (struct sockaddr*) &client_addr, &client_size ) ) == -1 )
	{
		module_master::ptr().broadcast( "*module_net_tcp: Clientverbindung konnte nicht akzeptiert werden.\n" );
		return( -1 );
	}

	module_master::ptr().broadcast( "!module_net_tcp: Verbindung entgegengenommen.\n" );

	is_connected = true;	
	is_server = true;
	return( 0 );
}

/// Stellt eine Verbindung zu einem Server her, der auf dem angegebenen
/// Port lauscht. Fr den praktischen Einsatz empfiehlt sich die Verwendung
/// eines hohen Ports, da die Ports von 1 - 1024 oftmals nur fr 
/// root verwendbar sind.
///
/// @param server Die Adresse des Servers. Kann eine IP-Adresse oder ein Hostname sein.
/// @param port Der Port, an dem der Server eingehende Verbindungen entgegennimmt.
///
/// @return 0, wenn die Verbindung hergestellt werden konnte. -1 sonst.
int module_net_tcp::connect( const char* server, unsigned int port )
{
	s = socket( PF_INET, SOCK_STREAM, 0 );
	if( inet_addr( server ) == INADDR_NONE )
	{
		// Auflsen des Namens 
		hostent* host = gethostbyname( server );
		if( host == NULL )
		{
			// TODO
			// Fehlermeldung oder nicht? Wenn dieser Fall auftritt,
			// stimmt eh irgendwas nicht...
			
			return( -1 );
		}

		sock_addr.sin_addr.s_addr = ((in_addr*)( host->h_addr ) )->s_addr;
	}
	else
		sock_addr.sin_addr.s_addr = inet_addr( server );
	

	// Socketdaten setzen

	sock_addr.sin_family = AF_INET;
	sock_addr.sin_port = htons( port );

	// Verbindung herstellen
		
	if( ::connect( s, (const sockaddr*) &sock_addr, sizeof( sock_addr ) ) == 0 )
	{
		is_connected = true;
		is_server = false;
		module_master::ptr().broadcast( "!module_net_tcp: Verbindung hergestellt.\n" );
		return( 0 );
	}
	
	module_master::ptr().broadcast( "#module_net_tcp: Keine Verbindung zum Server mglich.\n" );
	return( -1 );
}

/// Schliet die Verbindung oder beendet den Server,
/// je nachdem, welcher Modus gerade aktiv ist.
///
/// @return 0.
int module_net_tcp::disconnect( void )
{
	if( is_connected && !is_server )
	{
		is_connected = false;
		return( close( s ) );
	}

	is_server = false;
	return( close( c ) );
}

/// Sendet Nachrichten an den Server oder die Clients. Standardmig
/// werden dabei keine send flags gesetzt. In bestimmten Fllen
/// mag dies aber von Vorteil sein.
///
/// @param msg Die zu sendende Nachricht. 
/// @param size Die Gre der Nachricht in Bytes. 
/// @param flags Optionale Parameter. Unter send(2) finden sich mehr Informationen.
///
/// @return Anzahl der gesendeten Bytes (>= 0) oder -1, wenn ein Fehler aufgetreten ist. 
int module_net_tcp::send_msg( const char* msg, unsigned int size, int flags )
{
	if( msg == NULL )
		return( -1 );

	ssize_t res = 0;	

	if( !is_server && is_connected )
	{
		res = send( s, msg, size, flags );
		if( res < size )
			module_master::ptr().broadcast( "*module_net_tcp: Konnte Nachricht nicht senden.\n" );
		
		return( res );
	}
	else if( is_server )
	{
		// berprfe, ob es einen Fehler gab. Falls ja, warten wir von neuem
		// auf eine Verbindung. Es bleibt abzuwarten, ob das eine
		// gute Idee ist...
		if( (res = send( c, msg, size, flags )) == -1  )
		{
			is_connected = false;
			module_master::ptr().broadcast( "*module_net_tcp: Konnte Nachricht nicht senden. Warte auf neuen Client.\n" );
			
			close( c );
			socklen_t client_size = sizeof( struct sockaddr_in );
			memset( &client_addr, 0, client_size );
			
			if( ( c = accept( s, (struct sockaddr*) &client_addr, &client_size ) ) == -1 )
			{
				module_master::ptr().broadcast( "*module_net_tcp: Clientverbindung konnte nicht akzeptiert werden.\n" );
				return( -1 );
			}
	
			module_master::ptr().broadcast( "!module_net_tcp: Verbindung entgegengenommen.\n" );
			is_connected = true;
		}
		
		return( res );
	}

	return( -1 );	
}

/// Empfngt Nachrichten, die auf den Server oder den Client
/// warten.
///
/// @param buffer Der Puffer, in den die Daten geschrieben werden sollen. 
/// @param size Die Gre des Puffers in Bytes. 
/// @param flags Optionale flags. Unter recv(2) gibt es mehr Informationen.
///
/// @return Anzahl der empfangenen Bytes (>= 0) oder -1 im Fehlerfall.
int module_net_tcp::recv_msg( char* buffer, unsigned int size, int flags )
{
	if( buffer == NULL )
		return( -1 );

	if( has_msg() <= 0 )
		return( 0 );
	
	if( !is_server && is_connected )
		return( recv( s, buffer, size, flags ) );
	else
		return( recv( c, buffer, size, flags ) );
}

/// berprft, wie viele Bytes auf die Abholung am Socket warten.
///
/// @return Anzahl der Bytes, die auf Abholung warten oder -1, wenn ein Fehler
/// aufgetreten ist.
int module_net_tcp::has_msg( void )
{
	if( s == -1 || ( !is_connected && !is_server ) )
		return( -1 );

	int bytes = 0;

	if( !is_server )
		ioctl( s, FIONREAD, &bytes );
	else
		ioctl( c, FIONREAD, &bytes );	
	
	return( bytes );
}

/// Gibt den Status der Verbindung zurck.
///
/// @return true, wenn eine Verbindung vorliegt. false, wenn dem
/// nicht so ist. Dabei werden auch aktive Clientverbindungen
/// mit einbezogen.
bool module_net_tcp::has_connection( void )
{
	return( is_connected );
}

/// Gibt den aktuellen Status des Moduls zurck. Momentan nicht verwendet.
///
/// @return 0.
int module_net_tcp::get_status( void )
{
	return( 0 );
}
